from typing import Any
from triton._C.libproton import proton as libproton
import triton.runtime.driver as driver
import triton.language as tl
import triton
from triton import MockTensor
from .state import exit_state, enter_state, COMPUTE_METADATA_SCOPE_NAME


@triton.jit
def tensor_metric_kernel(device_ptr, device_offset_ptr, size: tl.uint64, metric_id: tl.uint64, metric_value_ptr,
                         metric_value_size: tl.uint64):
    BLOCK_SIZE: tl.constexpr = 128
    device_offset = tl.load(device_offset_ptr)
    tl.store(device_ptr + device_offset, metric_id)
    device_offset = (device_offset + 1) % size
    num_iters = tl.cdiv(metric_value_size, BLOCK_SIZE)
    offsets = tl.arange(0, BLOCK_SIZE)
    for i in tl.range(0, num_iters):
        cur_offsets = offsets + i * BLOCK_SIZE
        mask = cur_offsets < metric_value_size
        metric_value = tl.load(metric_value_ptr + cur_offsets, mask=mask)
        tl.store(device_ptr + (device_offset + cur_offsets) % size, metric_value, mask=mask)
    tl.debug_barrier()
    device_offset = (device_offset + metric_value_size) % size
    tl.store(device_offset_ptr, device_offset)


@triton.jit
def scalar_metric_kernel(device_ptr, device_offset_ptr, size: tl.uint64, metric_id: tl.uint64, metric_value: tl.uint64):
    device_offset = tl.load(device_offset_ptr)
    tl.store(device_ptr + device_offset, metric_id)
    device_offset = (device_offset + 1) % size
    tl.store(device_ptr + device_offset, metric_value)
    device_offset = (device_offset + 1) % size
    tl.debug_barrier()
    tl.store(device_offset_ptr, device_offset)


def _get_kernel(kernel_fn, *args):
    kernel = kernel_fn.warmup(*args, grid=(1, ), num_warps=1)
    kernel._init_handles()
    target = getattr(kernel.metadata, "target", None)
    warp_size = getattr(target, "warp_size", None)
    if warp_size is None:
        warp_size = driver.active.get_current_target().warp_size
    num_threads = kernel.metadata.num_warps * warp_size
    return kernel.function, num_threads, kernel.metadata.shared


def set_metric_kernels():
    mock_ptr = MockTensor(tl.uint64)
    mock_metric_id = 0
    mock_size = 1
    mock_metric_value_size = 1
    tensor_metric_kernel_fn, tensor_metric_kernel_num_threads, tensor_metric_kernel_shared = _get_kernel(
        tensor_metric_kernel,
        mock_ptr,
        mock_ptr,
        mock_size,
        mock_metric_id,
        mock_ptr,
        mock_metric_value_size,
    )
    scalar_metric_kernel_fn, scalar_metric_kernel_num_threads, scalar_metric_kernel_shared = _get_kernel(
        scalar_metric_kernel,
        mock_ptr,
        mock_ptr,
        mock_size,
        mock_metric_id,
        mock_metric_id,
    )
    device = driver.active.get_current_device()
    stream = driver.active.get_current_stream(device)
    libproton.set_metric_kernels(
        tensor_metric_kernel_fn,
        scalar_metric_kernel_fn,
        stream,
        tensor_metric_kernel_num_threads,
        tensor_metric_kernel_shared,
        scalar_metric_kernel_num_threads,
        scalar_metric_kernel_shared,
    )


class _TensorMetric(libproton.TensorMetric):

    def __init__(self, value, metric_type_index):
        super().__init__(value.data_ptr(), metric_type_index, value.numel())
        # Hold a reference to the backing tensor so its device memory stays alive.
        self._value = value


def transform_tensor_metrics(metrics: dict[str, Any]) -> tuple[dict[str, Any], dict[str, libproton.TensorMetric]]:
    tensor_metrics = {}
    scalar_metrics: dict[str, Any] = {}
    for key, value in metrics.items():
        if hasattr(value, "data_ptr"):  # tensor
            if value.device.type == "cpu":
                scalar_metrics[key] = value
            else:  # device tensor
                enter_state(COMPUTE_METADATA_SCOPE_NAME)
                # implicit casting to double or int64 tensors
                if value.is_floating_point():
                    value = value.double()
                    if value.numel() > 1:
                        metric_index = libproton.metric_type_vector_double_index
                    else:
                        metric_index = libproton.metric_type_double_index
                else:
                    value = value.long()
                    if value.numel() > 1:
                        metric_index = libproton.metric_type_vector_int64_index
                    else:
                        metric_index = libproton.metric_type_int64_index
                exit_state()
                tensor_metrics[key] = _TensorMetric(value, metric_index)
        else:
            scalar_metrics[key] = value
    return scalar_metrics, tensor_metrics



"""
Test module for proton's CPP API functionality.
No GPU kernel should be declared in this test.
Python API correctness tests involving GPU kernels should be placed in `test_api.py`.
Profile correctness tests involving GPU kernels should be placed in `test_profile.py`.
"""
import pathlib
import pytest

import triton._C.libproton.proton as libproton
from triton.profiler.profile import _select_backend


def test_record():
    id0 = libproton.record_scope()
    id1 = libproton.record_scope()
    assert id1 == id0 + 1


def test_state():
    libproton.enter_state("zero")
    libproton.exit_state()


def test_scope():
    id0 = libproton.record_scope()
    libproton.enter_scope(id0, "zero")
    id1 = libproton.record_scope()
    libproton.enter_scope(id1, "one")
    libproton.exit_scope(id1, "one")
    libproton.exit_scope(id0, "zero")


def test_op():
    id0 = libproton.record_scope()
    libproton.enter_op(id0, "zero")
    libproton.exit_op(id0, "zero")


@pytest.mark.parametrize("source", ["shadow", "python"])
def test_context(source: str, tmp_path: pathlib.Path):
    temp_file = tmp_path / "test_context.hatchet"
    session_id = libproton.start(str(temp_file.with_suffix("")), source, "tree", _select_backend())
    depth = libproton.get_context_depth(session_id)
    libproton.finalize(session_id, "hatchet")
    assert depth >= 0
    assert temp_file.exists()


def test_session(tmp_path: pathlib.Path):
    temp_file = tmp_path / "test_session.hatchet"
    session_id = libproton.start(str(temp_file.with_suffix("")), "shadow", "tree", _select_backend())
    libproton.deactivate(session_id, False)
    libproton.activate(session_id)
    libproton.finalize(session_id, "hatchet")
    libproton.finalize_all("hatchet")
    assert temp_file.exists()


def test_add_metrics(tmp_path: pathlib.Path):
    temp_file = tmp_path / "test_add_metrics.hatchet"
    libproton.start(str(temp_file.with_suffix("")), "shadow", "tree", _select_backend())
    id1 = libproton.record_scope()
    libproton.enter_scope(id1, "one")
    libproton.add_metrics(id1, {"a": 1.0, "b": 2.0, "veci": [1, 2, 3], "vecd": [1.0, 2.0, 3.0]})
    libproton.exit_scope(id1, "one")
    libproton.finalize_all("hatchet")
    assert temp_file.exists()


def test_init_function_metadata(tmp_path: pathlib.Path):
    metadata_file = tmp_path / "meta.json"
    metadata_file.write_text("{}")
    libproton.init_function_metadata(
        0,
        "dummy_fn",
        [(0, "root")],
        [],
        str(metadata_file),
    )
    libproton.destroy_function_metadata(0)


def test_instrumented_op_entry_exit():
    libproton.enter_instrumented_op(0, 0, 0, 0)
    libproton.exit_instrumented_op(0, 0, 0, 0)


def test_set_metric_kernels():
    libproton.set_metric_kernels(0, 0, 0)
    libproton.set_metric_kernels(0, 0, 0, 1, 0, 1, 0)


def test_tensor_metric_construction():
    metric = libproton.TensorMetric(123, libproton.metric_type_double_index)
    assert metric.ptr == 123
    assert metric.index == libproton.metric_type_double_index



from typing import Optional
from triton._C.libproton import proton as libproton  # type: ignore
import json as json
from .flags import flags


def get(session: Optional[int] = 0, phase: int = 0):
    """
    Retrieves profiling data for a given session.

    Args:
        session (Optional[int]): The session ID of the profiling session, or None if profiling is inactive.
    Returns:
        str: The profiling data in JSON format.
    """
    if session is None:
        return None
    if flags.command_line and session != 0:
        raise ValueError("Only one session can be retrieved when running from the command line.")
    return json.loads(libproton.get_data(session, phase))


def get_msgpack(session: Optional[int] = 0, phase: int = 0):
    """
    Retrieves profiling data for a given session encoded with MessagePack.

    Args:
        session (Optional[int]): The session ID of the profiling session, or None if profiling is inactive.

    Returns:
        bytes: The profiling data encoded with MessagePack.
    """
    if session is None:
        return None
    if flags.command_line and session != 0:
        raise ValueError("Only one session can be retrieved when running from the command line.")
    return libproton.get_data_msgpack(session, phase)


def advance_phase(session: Optional[int] = 0) -> Optional[int]:
    """
    Advances the profiling phase for a given session.

    Args:
        session (Optional[int]): The session ID of the profiling session, or None if profiling is inactive.

    Returns:
        Optional[int]: The next phase number after advancing.
    """
    if session is None:
        return None
    if flags.command_line and session != 0:
        raise ValueError("Only one session can advance phase when running from the command line.")
    return libproton.advance_data_phase(session)


def is_phase_complete(session: Optional[int] = 0, phase: int = 0) -> bool:
    """
    Checks if the profiling data for a given session and phase is complete.

    A "complete" phase is safe to read/clear because all device-side records for
    the phase have been flushed to the host and the phase will no longer receive
    new records.

    Args:
        session (Optional[int]): The session ID of the profiling session, or None if profiling is inactive.
        phase (int): The phase number to check. Defaults to 0.

    Returns:
        bool: True if the phase data is complete, False otherwise.
    """
    if session is None:
        return False
    if flags.command_line and session != 0:
        raise ValueError("Only one session can check phase completion status when running from the command line.")
    return libproton.is_data_phase_complete(session, phase)


def clear(
    session: Optional[int] = 0,
    phase: int = 0,
    clear_up_to_phase: bool = False,
) -> None:
    """
    Clears profiling data for a given session.

    Args:
        session (Optional[int]): The session ID of the profiling session, or None if profiling is inactive.
        phase (int): The phase number to clear. Defaults to 0.
        clear_up_to_phase (bool): If True, clear all phases up to and including `phase`.
    """
    if session is None:
        return
    if flags.command_line and session != 0:
        raise ValueError("Only one session can be cleared when running from the command line.")
    libproton.clear_data(session, phase, clear_up_to_phase)



"""
Intra-Kernel Profiling Examples using Proton DSL for Triton and Gluon Kernels
"""

import argparse

import torch
import triton
import triton.language as tl
import triton.profiler as proton
import triton.profiler.language as pl
from triton.experimental import gluon
from triton.experimental.gluon import language as gl
from triton.experimental.gluon.language.nvidia.hopper import (
    mbarrier,
    tma,
    warpgroup_mma,
    warpgroup_mma_init,
    warpgroup_mma_wait,
)

from triton.experimental.gluon.nvidia.hopper import TensorDescriptor

DEVICE = triton.runtime.driver.active.get_active_torch_device()

NUM_WARPS = 8


def is_hopper():
    target = triton.runtime.driver.active.get_current_target()
    return target.backend == "cuda" and torch.cuda.get_device_capability()[0] == 9


def config_helper(description: str):
    # Configure command line arguments for profiling options
    parser = argparse.ArgumentParser(description=description)
    parser.add_argument(
        "--op-measure",
        action="store_true",
        default=False,
        help="Enable operation measurement. Otherwise, we profile timeline trace. (default: False)",
    )
    parser.add_argument(
        "--warp-sampling",
        action="store_true",
        default=False,
        help="Enable warp sampling during profiling (default: False)",
    )
    parser.add_argument(
        "--increase-accuracy",
        action="store_true",
        default=False,
        help="Enable increased-accuracy during profiling (default: False).",
    )
    parser.add_argument(
        "--warp-ids",
        type=str,
        default="0, 2",
        help="Comma-separated list of warp IDs for warp sampling (default: '0, 2')",
    )
    parser.add_argument(
        "--gmem_buffer",
        action="store_true",
        default=False,
        help="Use global memory as the internal buffer during profiling (default: False).",
    )

    args = parser.parse_args()

    # Configure profiling options based on accuracy requirements
    # Default uses clock_64 for long-running kernels with higher overhead
    opts = ""
    # `clock_32` provides lower overhead per record, `time_shift`` post-processes to reduce noise
    if args.increase_accuracy:
        opts = "clock32,time_shift"

    if args.gmem_buffer:
        buf = "global"
    else:
        buf = "shared"

    # Set up profiling mode based on warp sampling preferences
    if args.warp_sampling:
        # Selective warp sampling allows capturing more events within buffer constraints
        # by only profiling specified warps (e.g. "0,1,2,3")
        mode = proton.mode.Default(
            optimizations=opts,
            sampling_strategy="selective",
            sampling_options=args.warp_ids,
            buffer_type=buf,
        )
    else:
        # Profile all warps - provides complete picture but uses more buffer space
        mode = proton.mode.Default(optimizations=opts, buffer_type=buf)

    return args.op_measure, mode


@triton.jit
def add_kernel(x_ptr,  # *Pointer* to first input vector.
               y_ptr,  # *Pointer* to second input vector.
               output_ptr,  # *Pointer* to output vector.
               n_elements,  # Size of the vector.
               BLOCK_SIZE: tl.constexpr,  # Number of elements each program should process.
               # NOTE: `constexpr` so it can be used as a shape value.
               ):
    pl.enter_scope("kernel")
    pid = tl.program_id(axis=0)
    block_start = pid * BLOCK_SIZE
    offsets = block_start + tl.arange(0, BLOCK_SIZE)
    mask = offsets < n_elements
    with pl.scope("load_and_add"):
        with pl.scope("load_x_issue"):
            x = tl.load(x_ptr + offsets, mask=mask)
        with pl.scope("load_y_issue"):
            y = tl.load(y_ptr + offsets, mask=mask)
        output = x + y
    tl.store(output_ptr + offsets, output, mask=mask)
    pl.exit_scope("kernel")


def add(x: torch.Tensor, y: torch.Tensor):
    output = torch.empty_like(x)
    assert x.device == DEVICE and y.device == DEVICE and output.device == DEVICE
    n_elements = output.numel()
    grid = lambda meta: (triton.cdiv(n_elements, meta["BLOCK_SIZE"]), )
    add_kernel[grid](x, y, output, n_elements, BLOCK_SIZE=1024, num_warps=NUM_WARPS)
    return output


if __name__ == "__main__":
    description = "Triton Vector Add with Proton Intra-Kernel Profiling"
    print(description)

    # Explicit Proton DSL enablement for Triton kernels.
    # Be careful NOT to insert proton ops in loops (use the ttgir override approach instead).
    pl.enable_semantic("triton")

    op_measure, mode = config_helper(description)

    # Start profiling with appropriate backend and output format
    if op_measure:
        # Operation measurement mode generates scope-level metrics
        # View results with: proton-viewer -m normalized_cycles vector-add.hatchet
        # Note: cycles are averaged across all warps/CTAs - adjust for warp specialization
        proton.start("vector-add", backend="instrumentation", mode=mode)
    else:
        # Timeline trace mode generates Chrome trace format for visualization
        # Output file: vector-add.chrome_trace
        proton.start("vector-add", data="trace", backend="instrumentation", mode=mode)

    torch.manual_seed(0)
    size = 98432
    x = torch.rand(size, device=DEVICE)
    y = torch.rand(size, device=DEVICE)
    output_torch = x + y
    output_triton = add(x, y)
    torch.testing.assert_close(output_torch, output_triton, rtol=1e-3, atol=1e-1)
    proton.finalize()


# This decorator allows us to invoke the function from a Gluon constexpr.
@gluon.constexpr_function
def get_warps_per_cta(BLOCK_M, BLOCK_N, num_warps):
    warps_per_cta = [4, 1]
    m = 16
    # Tile the atom until we have enough warps.
    while warps_per_cta[0] * warps_per_cta[1] != num_warps:
        # Tile along M only if it would not cause broadcasting.
        if BLOCK_M > m * warps_per_cta[0]:
            warps_per_cta[0] *= 2
        else:
            warps_per_cta[1] *= 2
    return warps_per_cta


@gluon.constexpr_function
def get_instr_shape_n(BLOCK_M, BLOCK_N, num_warps):
    m = 16
    mReps = triton.cdiv(BLOCK_M, m)
    nReps = triton.cdiv(num_warps, mReps)
    maxN = max(BLOCK_N // nReps, 8)
    n = 256
    while n > maxN or BLOCK_N % n != 0:
        n -= 8
    assert n >= 8, "expected to find a valid n"
    return n


@gluon.constexpr_function
def pick_wgmma_layout(dtype, BLOCK_M, BLOCK_N, num_warps):
    m = 16
    k = 256 // dtype.primitive_bitwidth
    n = get_instr_shape_n(BLOCK_M, BLOCK_N, num_warps)
    warps_per_cta = get_warps_per_cta(BLOCK_M, BLOCK_N, num_warps)
    return gl.NVMMADistributedLayout(
        version=[3, 0],
        warps_per_cta=warps_per_cta,
        instr_shape=[m, n, k],
    )


@gluon.jit
def blocked_matmul_pipelined_kernel(a_desc, b_desc, c_desc, num_warps: gl.constexpr):
    BLOCK_M: gl.constexpr = c_desc.block_type.shape[0]
    BLOCK_N: gl.constexpr = c_desc.block_type.shape[1]
    BLOCK_K: gl.constexpr = a_desc.block_type.shape[1]
    dtype: gl.constexpr = a_desc.dtype
    K = a_desc.shape[1]

    pl.enter_scope("blocked_matmul_pipelined_kernel")

    # Allocate 2 buffers for each A and B.
    a_smem = gl.allocate_shared_memory(dtype, [2] + a_desc.block_type.shape, a_desc.layout)
    b_smem = gl.allocate_shared_memory(dtype, [2] + b_desc.block_type.shape, b_desc.layout)
    index = 0

    pid_m = gl.program_id(axis=0)
    pid_n = gl.program_id(axis=1)
    off_m = pid_m * BLOCK_M
    off_n = pid_n * BLOCK_N

    mma_layout: gl.constexpr = pick_wgmma_layout(dtype, BLOCK_M, BLOCK_N, num_warps)
    acc = warpgroup_mma_init(gl.zeros((BLOCK_M, BLOCK_N), dtype=gl.float32, layout=mma_layout))

    bar = gl.allocate_shared_memory(gl.int64, [1], mbarrier.MBarrierLayout())
    mbarrier.init(bar, count=1)
    phase = 0

    for k in range(0, K, BLOCK_K):
        a = a_smem.index(index)
        b = b_smem.index(index)

        mbarrier.expect(bar, a_desc.block_type.nbytes + b_desc.block_type.nbytes)

        with pl.scope("tma_loads_issue"):
            tma.async_copy_global_to_shared(a_desc, [off_m, k], bar, a)
            tma.async_copy_global_to_shared(b_desc, [k, off_n], bar, b)

        with pl.scope("tma_loads_wait"):
            mbarrier.wait(bar, phase=phase)
        phase ^= 1

        # Since `warpgroup_mma_wait` is a no-op when there are no WGMMAs in
        # flight, we can overlap the WGMMA by waiting first, then issuing the
        # async WGMMA.
        with pl.scope("wgmma_wait"):
            acc = warpgroup_mma_wait(num_outstanding=0, deps=(acc, ))

        with pl.scope("wgmma_issue"):
            acc = warpgroup_mma(a, b, acc, is_async=True)

        # Move to the next buffer. The TMA load will start while the WGMMA is
        # still running.
        index ^= 1

    # Wait for the last WGMMA to complete.
    with pl.scope("wgmma_last_wait"):
        acc = warpgroup_mma_wait(num_outstanding=0, deps=(acc, ))

    mbarrier.invalidate(bar)

    c_smem = gl.allocate_shared_memory(dtype, c_desc.block_type.shape, c_desc.layout)
    c_smem.store(acc.to(dtype))
    tma.async_copy_shared_to_global(c_desc, [off_m, off_n], c_smem)
    tma.store_wait(pendings=0)

    pl.exit_scope("blocked_matmul_pipelined_kernel")


def blocked_matmul_pipelined(A, B, C, BLOCK_M, BLOCK_N, BLOCK_K, num_warps):
    M, N = C.shape

    a_layout = gl.NVMMASharedLayout.get_default_for([BLOCK_M, BLOCK_K], gl.float16)
    b_layout = gl.NVMMASharedLayout.get_default_for([BLOCK_K, BLOCK_N], gl.float16)
    c_layout = gl.NVMMASharedLayout.get_default_for([BLOCK_M, BLOCK_N], gl.float16)
    a_desc = TensorDescriptor.from_tensor(A, [BLOCK_M, BLOCK_K], a_layout)
    b_desc = TensorDescriptor.from_tensor(B, [BLOCK_K, BLOCK_N], b_layout)
    c_desc = TensorDescriptor.from_tensor(C, [BLOCK_M, BLOCK_N], c_layout)

    grid = (triton.cdiv(M, BLOCK_M), triton.cdiv(N, BLOCK_N))
    blocked_matmul_pipelined_kernel[grid](a_desc, b_desc, c_desc, num_warps=num_warps)


if __name__ == "__main__":
    if not is_hopper():
        raise RuntimeError("This tutorial requires a Hopper NVIDIA GPU")

    description = "Gluon Matrix Multiplication with Proton Intra-Kernel Profiling"
    print(description)

    M, N, K = 512, 512, 1024
    A = torch.randn(M, K, device="cuda", dtype=torch.float16)
    B = torch.randn(K, N, device="cuda", dtype=torch.float16)
    C = torch.empty(M, N, device="cuda", dtype=torch.float16)
    BLOCK_M, BLOCK_N, BLOCK_K = 128, 128, 128

    op_measure, mode = config_helper(description)

    # Start profiling with appropriate backend and output format
    if op_measure:
        # Operation measurement mode generates scope-level metrics
        # View results with: proton-viewer -m normalized_cycles gemm.hatchet
        # Note: cycles are averaged across all warps/CTAs - adjust for warp specialization
        proton.start("gemm", backend="instrumentation", mode=mode)
    else:
        # Timeline trace mode generates Chrome trace format for visualization
        # Output file: gemm.chrome_trace
        proton.start("gemm", data="trace", backend="instrumentation", mode=mode)

    blocked_matmul_pipelined(A, B, C, BLOCK_M, BLOCK_N, BLOCK_K, NUM_WARPS)
    torch.testing.assert_close(A @ B, C, rtol=1e-3, atol=1e-1)

    # Complete profiling and write output files
    proton.finalize()



from ..state import enter_state, exit_state, COMPUTE_METADATA_SCOPE_NAME
from ..metric import transform_tensor_metrics, set_metric_kernels
from triton.compiler import LazyDict
from .hook import Hook
from triton._C.libproton import proton as libproton
from contextvars import ContextVar
from numbers import Number
import re
from typing import Optional

op_name = ContextVar("op_name", default=None)
id = ContextVar("id", default=None)
enabled = ContextVar("enabled", default=False)


class LaunchHook(Hook):
    # Highest priority
    priority = 100
    flops_width = [8, 16, 32, 64]
    # Historical/derived metrics (e.g., used by viewer utilization computations).
    # Launch metadata can carry *additional* metrics; see _extract_metrics().
    metrics = [f"flops{width}" for width in flops_width] + ["bytes"] + ["flops"]

    # Reserved keys that Triton’s runtime always attaches to launch_metadata.
    # We never treat these as metrics.
    _reserved_metadata_keys = {"name", "function", "stream"}

    # LaunchHook is intended to be a process-wide singleton. HookManager dedupes
    # by identity (object instance), so we must ensure repeated LaunchHook()
    # constructions return the same instance to avoid double registration.
    _instance = None

    def configure(self, *, include: Optional[str] = None, exclude: Optional[str] = None) -> None:
        # Regexes over the compiled kernel name (metadata.data["name"]).
        self._include_pattern = include
        self._exclude_pattern = exclude
        self._include_re = re.compile(include) if include else None
        self._exclude_re = re.compile(exclude) if exclude else None

    def _matches_kernel_name(self, kernel_name: str) -> bool:
        if self._include_re is not None and self._include_re.match(kernel_name) is None:
            return False
        if self._exclude_re is not None and self._exclude_re.match(kernel_name) is not None:
            return False
        return True

    @staticmethod
    def _is_supported_metric_value(value) -> bool:
        # Supported scalar: Python/numpy number-like (bools are allowed but not very useful).
        # Supported tensor: objects with a data_ptr() method (e.g., torch.Tensor).
        if value is None:
            return False
        if hasattr(value, "data_ptr"):
            return True
        return isinstance(value, Number)

    @staticmethod
    def _extract_metrics(lazy_metadata: dict) -> dict:
        # Accept arbitrary metrics from launch_metadata while filtering out reserved fields
        # and unsupported values (e.g., objects/functions).
        return {
            k: v
            for k, v in lazy_metadata.items()
            if k not in LaunchHook._reserved_metadata_keys and LaunchHook._is_supported_metric_value(v)
        }

    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
            cls._instance._initialized = False
        return cls._instance

    def __init__(self):
        # Singleton: __init__ is invoked on every construction even when __new__
        # returns an existing instance.
        if getattr(self, "_initialized", False):
            return
        # Ensure filter state is always initialized even if configure() isn't called.
        self.configure(include=None, exclude=None)
        self._initialized = True

    def init_handle(self, module, function, name: str, metadata_group: dict, hash: str) -> None:
        pass

    def activate(self):
        pass

    def deactivate(self):
        pass

    def enter(self, metadata: LazyDict) -> None:
        # Fast path: if the kernel name is already available without evaluating launch_metadata,
        # apply include/exclude filters and potentially skip metadata evaluation entirely.
        kernel_name = metadata.data.get("name")
        if not self._matches_kernel_name(kernel_name):
            enabled.set(False)
            return
        enter_state(COMPUTE_METADATA_SCOPE_NAME)
        lazy_metadata = metadata.get()
        exit_state()

        kernel_name = lazy_metadata["name"]
        # If name wasn't available (or changed), apply filters using the evaluated name.
        if not self._matches_kernel_name(kernel_name):
            enabled.set(False)
            return

        enabled.set(True)
        fn_metrics = LaunchHook._extract_metrics(lazy_metadata)
        op_name.set(kernel_name)
        id.set(libproton.record_scope())
        if fn_metrics:
            set_metric_kernels()
        scalar_metrics, tensor_metrics = transform_tensor_metrics(fn_metrics)
        libproton.enter_op(id.get(), lazy_metadata["name"])
        libproton.add_metrics(id.get(), scalar_metrics, tensor_metrics)

    def exit(self, metadata: LazyDict) -> None:
        if not enabled.get():
            return
        libproton.exit_op(id.get(), op_name.get())



import pytest
import subprocess
from triton.profiler.viewer import get_min_time_flops, get_min_time_bytes, read, format_frames, derive_metrics, filter_frames, parse, apply_diff_profile
from triton.profiler.hooks.launch import COMPUTE_METADATA_SCOPE_NAME
import numpy as np

file_path = __file__
triton_example_file = file_path.replace("test_viewer.py", "examples/triton.json")
cuda_example_file = file_path.replace("test_viewer.py", "examples/cuda.json")
hip_example_file = file_path.replace("test_viewer.py", "examples/hip.json")
frame_example_file = file_path.replace("test_viewer.py", "examples/frame.json")
leaf_example_file = file_path.replace("test_viewer.py", "examples/leaf_nodes.json")


def test_help():
    # Only check if the viewer can be invoked
    subprocess.check_call(["proton-viewer", "-h"], stdout=subprocess.DEVNULL)


def test_exclusive_metrics():
    gf, inclusive_metrics, exclusive_metrics, device_info = read(triton_example_file)
    metrics = ["cpu_time/ns"]
    metrics = derive_metrics(gf, metrics, inclusive_metrics, exclusive_metrics, device_info)
    gf = filter_frames(gf, None, None, None, metrics[0])
    sorted_df = gf.dataframe.sort_values(by=[metrics[0]], ascending=False)
    actual = sorted_df.iloc[0:1]["name"].values[0]
    assert actual == "scope"


def test_sort():
    gf, inclusive_metrics, exclusive_metrics, device_info = read(leaf_example_file)
    gf = format_frames(gf, None)
    metrics = ["time/s", "time/ms", "time/us", "time/ns"]
    metrics = derive_metrics(gf, metrics, inclusive_metrics, exclusive_metrics, device_info)
    gf = filter_frames(gf, None, None, None, metrics[0])
    sorted_df = gf.dataframe.sort_values(by=[metrics[0]], ascending=False)
    actual = sorted_df.iloc[0:5]["name"].values
    expected = ["ROOT", "kernel_1_1_1", "kernel_3_1_1", "kernel_3_2_2", "kernel_1_2_2"]
    assert len(actual) == len(expected)
    assert all(a == b for a, b in zip(actual, expected))


@pytest.mark.parametrize("option", ["full", "file_function_line", "function_line", "file_function"])
def test_format_frames(option):
    gf, _, _, _ = read(frame_example_file)
    gf = format_frames(gf, option)
    if option == "full":
        idx = gf.dataframe["name"] == "/home/user/projects/example.py/test.py:1@foo"
    elif option == "file_function_line":
        idx = gf.dataframe["name"] == "test.py:1@foo"
    elif option == "function_line":
        idx = gf.dataframe["name"] == "1@foo"
    elif option == "file_function":
        idx = gf.dataframe["name"] == "test.py@foo"
    assert idx.sum() == 1


@pytest.mark.parametrize("option", ["include", "exclude"])
def test_filter_frames(option):
    include = ""
    exclude = ""
    gf, _, _, _ = read(frame_example_file)
    if option == "include":
        include = ".*test0.*"
    elif option == "exclude":
        exclude = ".*test1.*"
    gf = filter_frames(gf, include=include, exclude=exclude)
    idx = gf.dataframe["name"] == "test1"
    assert idx.sum() == 0
    idx = gf.dataframe["name"] == "test0"
    assert idx.sum() == 1


def test_filter_metadata():
    gf, _, _, _ = read(triton_example_file)
    assert COMPUTE_METADATA_SCOPE_NAME not in gf.dataframe["name"].tolist()
    assert "cuda_kernel" not in gf.dataframe["name"].tolist()
    assert "scope" in gf.dataframe["name"].tolist()
    assert "triton_kernel" in gf.dataframe["name"].tolist()


def test_parse():
    gf, derived_metrics = parse(["time/s"], triton_example_file)
    for derived_metric in derived_metrics:
        assert derived_metric in gf.inc_metrics or derived_metric in gf.exc_metrics


def test_min_time_flops():
    gf, _, _, device_info = read(cuda_example_file)
    ret = get_min_time_flops(gf.dataframe, device_info)
    device0_idx = gf.dataframe["device_id"] == "0"
    device1_idx = gf.dataframe["device_id"] == "1"
    device2_idx = gf.dataframe["device_id"] == "2"
    # sm89
    np.testing.assert_allclose(ret[device0_idx].to_numpy(), [[0.000025]], atol=1e-5)
    # sm90
    np.testing.assert_allclose(ret[device1_idx].to_numpy(), [[0.00005]], atol=1e-5)
    # sm100
    np.testing.assert_allclose(ret[device2_idx].to_numpy(), [[0.000025]], atol=1e-5)
    gf, _, _, device_info = read(hip_example_file)
    ret = get_min_time_flops(gf.dataframe, device_info)
    device0_idx = gf.dataframe["device_id"] == "0"
    device1_idx = gf.dataframe["device_id"] == "1"
    device2_idx = gf.dataframe["device_id"] == "2"
    # CDNA2
    np.testing.assert_allclose(ret[device0_idx].to_numpy(), [[0.000055]], atol=1e-5)
    # CDNA3
    np.testing.assert_allclose(ret[device1_idx].to_numpy(), [[0.000038]], atol=1e-5)
    # CDNA4
    np.testing.assert_allclose(ret[device2_idx].to_numpy(), [[0.000217]], atol=1e-5)


def test_min_time_bytes():
    gf, _, _, device_info = read(cuda_example_file)
    ret = get_min_time_bytes(gf.dataframe, device_info)
    device0_idx = gf.dataframe["device_id"] == "0"
    device1_idx = gf.dataframe["device_id"] == "1"
    # sm89
    np.testing.assert_allclose(ret[device0_idx].to_numpy(), [[9.91969e-06]], atol=1e-6)
    # sm90
    np.testing.assert_allclose(ret[device1_idx].to_numpy(), [[2.48584e-05]], atol=1e-6)
    gf, _, _, device_info = read(hip_example_file)
    ret = get_min_time_bytes(gf.dataframe, device_info)
    device0_idx = gf.dataframe["device_id"] == "0"
    device1_idx = gf.dataframe["device_id"] == "1"
    device2_idx = gf.dataframe["device_id"] == "2"
    # CDNA2
    np.testing.assert_allclose(ret[device0_idx].to_numpy(), [[3.125e-06]], atol=1e-6)
    # CDNA3
    np.testing.assert_allclose(ret[device1_idx].to_numpy(), [[1.93378e-05]], atol=1e-6)
    # CDNA4
    np.testing.assert_allclose(ret[device2_idx].to_numpy(), [[0.000125]], atol=1e-6)


def test_percentage():
    pass


def derivation_metrics_test(metrics, expected_data, sample_file, rtol=1e-7, atol=1e-6):
    gf, inclusive_metrics, exclusive_metrics, device_info = read(sample_file)
    assert len(inclusive_metrics + exclusive_metrics) > 0, "No metrics found in the input file"
    derived_metrics = derive_metrics(gf, metrics, inclusive_metrics, exclusive_metrics, device_info)
    for derived_metric in derived_metrics:
        np.testing.assert_allclose(gf.dataframe[derived_metric].to_numpy(), expected_data[derived_metric], rtol=rtol,
                                   atol=atol)


def test_avg_time_derivation():
    derivation_metrics_test(
        metrics=["avg_time/s", "avg_time/ms", "avg_time/us", "avg_time/ns"], expected_data={
            "avg_time/s (inc)": [0.0000512, 0.0000205, 0.000205,
                                 0.000205], "avg_time/ms (inc)": [0.0512, 0.02048, 0.2048, 0.2048], "avg_time/us (inc)":
            [51.2, 20.48, 204.8, 204.8], "avg_time/ns (inc)": [51200.0, 20480.0, 204800.0, 204800.0]
        }, sample_file=cuda_example_file)


def test_util():
    derivation_metrics_test(metrics=["util"], expected_data={
        "util": [np.nan, 0.247044, 0.147830, 0.118451],
    }, sample_file=cuda_example_file)


def test_time_derivation():
    derivation_metrics_test(
        metrics=["time/s", "time/ms", "time/us", "time/ns"], expected_data={
            "time/s (inc)": [0.000614, 0.0002048, 0.0002048, 0.0002048],
            "time/ms (inc)": [0.6144, 0.2048, 0.2048, 0.2048],
            "time/us (inc)": [614.4, 204.8, 204.8, 204.8],
            "time/ns (inc)": [614400.0, 204800.0, 204800.0, 204800.0],
            "time/% (inc)": [100.0, 50.0, 50.0, 50.0],
        }, sample_file=cuda_example_file)


def test_bytes_derivation():
    derivation_metrics_test(
        metrics=["byte/s", "gbyte/s", "tbyte/s"], expected_data={
            "byte/s (inc)": [1.953125e+11, 4.88281250e+11, 4.88281250e+10,
                             4.88281250e+10], "gbyte/s (inc)": [195.3125, 488.28125, 48.828125, 48.828125],
            "tbyte/s (inc)": [0.195312, 0.48828125, 0.04882812, 0.04882812]
        }, sample_file=cuda_example_file)


def test_flops_derivation():
    derivation_metrics_test(
        metrics=["flop8/s", "gflop8/s", "tflop8/s"],
        expected_data={
            "flop8/s (inc)": [3.417969e+14, 4.88281250e+14, 4.88281250e+13,
                              4.88281250e+14], "gflop8/s (inc)": [341796.875, 488281.25, 48828.125, 488281.25],
            "tflop8/s (inc)": [341.796875, 488.28125, 48.828125, 488.28125]
        },
        sample_file=cuda_example_file,
    )


def test_diff_profile():
    gf, derived_metrics = parse(["time/s"], triton_example_file)
    gf2, _ = parse(["time/s"], cuda_example_file)
    gf = apply_diff_profile(gf, derived_metrics, cuda_example_file, ["time/s"], None, None, 0.0)
    assert "time/s (inc)" in gf.dataframe.columns



"""
Test module for proton's Python API.
No GPU kernel should be declared in this test.
Profile correctness tests involving GPU kernels should be placed in `test_profile.py`.
"""

import pytest
import json
import triton.profiler as proton
import pathlib
from triton.profiler.hooks.hook import HookManager
from triton.profiler.hooks.launch import LaunchHook
from triton.profiler.hooks.instrumentation import InstrumentationHook
from triton._internal_testing import is_hip


def test_profile_single_session(tmp_path: pathlib.Path):
    temp_file0 = tmp_path / "test_profile0.hatchet"
    session_id0 = proton.start(str(temp_file0.with_suffix("")))
    proton.activate()
    proton.deactivate()
    proton.finalize()
    assert temp_file0.exists()

    temp_file1 = tmp_path / "test_profile1.hatchet"
    session_id1 = proton.start(str(temp_file1.with_suffix("")))
    proton.activate(session_id1)
    proton.deactivate(session_id1)
    proton.finalize(session_id1)
    assert session_id1 == session_id0 + 1
    assert temp_file1.exists()

    session_id2 = proton.start("test")
    proton.activate(session_id2)
    proton.deactivate(session_id2)
    proton.finalize()
    assert session_id2 == session_id1 + 1
    assert pathlib.Path("test.hatchet").exists()
    pathlib.Path("test.hatchet").unlink()


def test_profile_multiple_sessions(tmp_path: pathlib.Path):
    temp_file0 = tmp_path / "test_profile0.hatchet"
    proton.start(str(temp_file0.with_suffix("")))
    temp_file1 = tmp_path / "test_profile1.hatchet"
    proton.start(str(temp_file1.with_suffix("")))
    proton.activate()
    proton.deactivate()
    proton.finalize()
    assert temp_file0.exists()
    assert temp_file1.exists()

    temp_file2 = tmp_path / "test_profile2.hatchet"
    session_id2 = proton.start(str(temp_file2.with_suffix("")))
    temp_file3 = tmp_path / "test_profile3.hatchet"
    session_id3 = proton.start(str(temp_file3.with_suffix("")))
    proton.deactivate(session_id2)
    proton.deactivate(session_id3)
    proton.finalize()
    assert temp_file2.exists()
    assert temp_file3.exists()


def test_profile_mode(tmp_path: pathlib.Path):
    temp_file0 = tmp_path / "test_profile0.hatchet"
    if is_hip():
        try:
            proton.start(str(temp_file0.with_suffix("")), mode="pcsampling")
        except Exception as e:
            assert "RoctracerProfiler: unsupported mode: pcsampling" in str(e)
        finally:
            proton.finalize()
    else:
        import os
        import pytest

        if os.environ.get("PROTON_SKIP_PC_SAMPLING_TEST", "0") == "1":
            pytest.skip("PC sampling test is disabled")

        # Two sessions with the same mode can coexist
        proton.start(str(temp_file0.with_suffix("")), mode="pcsampling")
        temp_file1 = tmp_path / "test_profile1.hatchet"
        proton.start(str(temp_file1.with_suffix("")), mode="pcsampling")
        proton.finalize()
        assert temp_file1.exists()

        # Two sessions with different modes cannot coexist
        try:
            proton.start(str(temp_file0.with_suffix("")), mode="pcsampling")
            proton.start(str(temp_file1.with_suffix("")))
        except Exception as e:
            assert "Cannot add a session with the same profiler but a different mode than existing sessions" in str(e)
        finally:
            proton.finalize()

        # Two sessions with different modes cannot coexist even if the first session is deactivated.
        # In proton, once we deactivate a session, its profiler is not stopped, so changing the profiler mode is not allowed
        # The only way to start a session with a different mode is to finalize all existing sessions first.
        try:
            session_id = proton.start(str(temp_file0.with_suffix("")), mode="pcsampling")
            proton.deactivate(session_id)
            temp_file1 = tmp_path / "test_profile1.hatchet"
            proton.start(str(temp_file1.with_suffix("")))
        except Exception as e:
            assert "Cannot add a session with the same profiler but a different mode than existing sessions" in str(e)
        finally:
            proton.finalize()


def test_profile_decorator(tmp_path: pathlib.Path):
    temp_file = tmp_path / "test_profile_decorator.hatchet"

    @proton.profile(name=str(temp_file.with_suffix("")))
    def foo0(a, b):
        return a + b

    foo0(1, 2)
    proton.finalize()
    assert temp_file.exists()

    @proton.profile
    def foo1(a, b):
        return a + b

    foo1(1, 2)
    proton.finalize()
    default_file = pathlib.Path(proton.DEFAULT_PROFILE_NAME + ".hatchet")
    assert default_file.exists()
    default_file.unlink()


def test_scope(tmp_path: pathlib.Path):
    # Scope can be annotated even when profiling is off
    with proton.scope("test"):
        pass

    temp_file = tmp_path / "test_scope.hatchet"
    proton.start(str(temp_file.with_suffix("")))
    with proton.scope("test"):
        pass

    @proton.scope("test")
    def foo():
        pass

    foo()

    proton.enter_scope("test")
    proton.exit_scope()

    proton.enter_scope("test0")
    proton.exit_scope("test0")

    proton.finalize()
    assert temp_file.exists()


def test_hook(tmp_path: pathlib.Path):
    temp_file = tmp_path / "test_hook.hatchet"
    session_id0 = proton.start(str(temp_file.with_suffix("")), hook="triton")
    proton.activate(session_id0)
    proton.activate(session_id0)
    assert len(
        HookManager.active_hooks) == 1, ("Activate a session multiple times should maintain a single instance of hook")
    assert list(HookManager.session_hooks[session_id0].values())[0] is True
    proton.deactivate(session_id0)
    assert list(HookManager.session_hooks[session_id0].values())[0] is False
    assert len(HookManager.active_hooks) == 0
    # Deactivate a session multiple times should not raise an error
    proton.deactivate(session_id0)
    proton.finalize(None)
    assert temp_file.exists()


def test_hook_manager(tmp_path: pathlib.Path):
    # Launch hook is a singleton
    HookManager.register(LaunchHook(), 0)
    HookManager.register(LaunchHook(), 0)
    assert len(HookManager.active_hooks) == 1
    assert isinstance(HookManager.active_hooks[0], LaunchHook)
    assert HookManager.session_hooks[0][HookManager.active_hooks[0]] is True

    # Only unregister one session
    HookManager.register(LaunchHook(), 1)
    HookManager.unregister(0)
    assert len(HookManager.active_hooks) == 1
    HookManager.unregister(1)
    assert len(HookManager.active_hooks) == 0

    # Heterogenous hooks
    HookManager.register(InstrumentationHook(""), 2)
    HookManager.register(LaunchHook(), 2)
    assert len(HookManager.active_hooks) == 2
    # Launch hook has a higher priority
    assert isinstance(HookManager.active_hooks[0], LaunchHook)
    assert isinstance(HookManager.active_hooks[1], InstrumentationHook)
    assert HookManager.session_hooks[2][HookManager.active_hooks[0]] is True
    assert HookManager.session_hooks[2][HookManager.active_hooks[1]] is True
    HookManager.unregister()
    assert len(HookManager.active_hooks) == 0


def test_scope_metrics(tmp_path: pathlib.Path):
    temp_file = tmp_path / "test_scope_metrics.hatchet"
    session_id = proton.start(str(temp_file.with_suffix("")))
    # Test different scope creation methods
    with proton.scope("test0", {"a": 1.0}):
        pass

    @proton.scope("test1", {"a": 1.0})
    def foo():
        pass

    foo()

    # After deactivation, the metrics should be ignored
    proton.deactivate(session_id)
    proton.enter_scope("test2", metrics={"a": 1.0})
    proton.exit_scope()

    # Metrics should be recorded again after reactivation
    proton.activate(session_id)
    proton.enter_scope("test3", metrics={"a": 1.0})
    proton.exit_scope()

    proton.enter_scope("test3", metrics={"a": 1.0})
    proton.exit_scope()

    # exit_scope can also take metrics
    proton.enter_scope("test4")
    proton.exit_scope(metrics={"b": 1.0})

    proton.finalize()

    assert temp_file.exists()
    with temp_file.open() as f:
        data = json.load(f)
    assert len(data[0]["children"]) == 4
    for child in data[0]["children"]:
        if child["frame"]["name"] == "test3":
            assert child["metrics"]["a"] == 2.0
        elif child["frame"]["name"] == "test4":
            assert child["metrics"]["b"] == 1.0


def test_scope_metrics_invalid(tmp_path: pathlib.Path):
    temp_file = tmp_path / "test_scope_metrics.hatchet"
    proton.start(str(temp_file.with_suffix("")))

    error = None

    try:
        with proton.scope("test0", {"a": 1.0}):
            pass

        with proton.scope("test0", {"a": 1}):
            pass
    except Exception as e:
        error = str(e)
    finally:
        proton.finalize()

    assert error is not None and "Metric value type mismatch for valueId 0 (a): current=double, new=int64_t" in error


def test_scope_properties(tmp_path: pathlib.Path):
    temp_file = tmp_path / "test_scope_properties.hatchet"
    proton.start(str(temp_file.with_suffix("")))
    # Properties do not aggregate
    proton.enter_scope("test0", metrics={"a (pty)": 1.0})
    proton.exit_scope()

    proton.enter_scope("test0", metrics={"a (pty)": 1.0})
    proton.exit_scope()

    proton.finalize()
    assert temp_file.exists()
    with temp_file.open() as f:
        data = json.load(f)
    for child in data[0]["children"]:
        if child["frame"]["name"] == "test0":
            assert child["metrics"]["a"] == 1.0


def test_scope_exclusive(tmp_path: pathlib.Path):
    temp_file = tmp_path / "test_scope_exclusive.hatchet"
    proton.start(str(temp_file.with_suffix("")))
    # metric a only appears in the outermost scope
    # metric b only appears in the innermost scope
    # both metrics do not appear in the root scope
    with proton.scope("test0", metrics={"a (exc)": 1}):
        with proton.scope("test1", metrics={"b (exc)": 1}):
            pass

    proton.finalize()
    assert temp_file.exists()
    with temp_file.open() as f:
        data = json.load(f)
    root_metrics = data[0]["metrics"]
    assert len(root_metrics) == 0
    test0_frame = data[0]["children"][0]
    test0_metrics = test0_frame["metrics"]
    assert len(test0_metrics) == 1
    assert test0_metrics["a"] == 1
    test1_frame = test0_frame["children"][0]
    test1_metrics = test1_frame["metrics"]
    assert len(test1_metrics) == 1
    assert test1_metrics["b"] == 1


def test_scope_vector_metrics(tmp_path: pathlib.Path):
    temp_file = tmp_path / "test_scope_vector_metrics.hatchet"
    proton.start(str(temp_file.with_suffix("")))
    proton.enter_scope("test0", metrics={"veci": [1, 2, 3], "vecd": [1.0, 2.0, 3.0]})
    proton.exit_scope()
    proton.finalize()
    assert temp_file.exists()
    with temp_file.open() as f:
        data = json.load(f)
    for child in data[0]["children"]:
        if child["frame"]["name"] == "test0":
            assert child["metrics"]["veci"] == [1, 2, 3]
            assert child["metrics"]["vecd"] == [1.0, 2.0, 3.0]


def test_state(tmp_path: pathlib.Path):
    temp_file = tmp_path / "test_state.hatchet"
    proton.start(str(temp_file.with_suffix("")))
    proton.enter_scope("test0")
    proton.enter_state("state")
    proton.enter_scope("test1", metrics={"a": 1.0})
    proton.exit_scope()
    proton.exit_state()
    proton.exit_scope()
    proton.finalize()
    assert temp_file.exists()
    with temp_file.open() as f:
        data = json.load(f)
    # test0->test1->state
    assert len(data[0]["children"]) == 1
    child = data[0]["children"][0]
    assert child["frame"]["name"] == "test0"
    assert len(child["children"]) == 1
    child = child["children"][0]
    assert child["frame"]["name"] == "test1"
    assert len(child["children"]) == 1
    child = child["children"][0]
    assert child["frame"]["name"] == "state"
    assert child["metrics"]["a"] == 1.0


def test_context_depth(tmp_path: pathlib.Path):
    temp_file = tmp_path / "test_context_depth.hatchet"
    session_id = proton.start(str(temp_file.with_suffix("")))
    assert proton.context.depth(session_id) == 0
    proton.enter_scope("test0")
    assert proton.context.depth(session_id) == 1
    proton.enter_scope("test1")
    assert proton.context.depth(session_id) == 2
    proton.exit_scope()
    assert proton.context.depth(session_id) == 1
    proton.exit_scope()
    assert proton.context.depth(session_id) == 0
    proton.finalize()


def test_throw(tmp_path: pathlib.Path):
    # Catch an exception thrown by c++
    session_id = 100
    temp_file = tmp_path / "test_throw.hatchet"
    activate_error = ""
    try:
        session_id = proton.start(str(temp_file.with_suffix("")))
        proton.activate(session_id + 1)
    except Exception as e:
        activate_error = str(e)
    finally:
        proton.finalize()
    assert "Session has not been initialized: " + str(session_id + 1) in activate_error

    deactivate_error = ""
    try:
        session_id = proton.start(str(temp_file.with_suffix("")))
        proton.deactivate(session_id + 1)
    except Exception as e:
        deactivate_error = str(e)
    finally:
        proton.finalize()
    assert "Session has not been initialized: " + str(session_id + 1) in deactivate_error


@pytest.mark.parametrize("disable", [True, False])
def test_profile_disable(disable, fresh_knobs, tmp_path: pathlib.Path):
    fresh_knobs.proton.disable = disable
    temp_file = tmp_path / "test_profile_disable.hatchet"
    proton.start(str(temp_file.with_suffix("")))
    proton.enter_scope("test0")
    proton.exit_scope()
    proton.finalize()
    if disable:
        assert not temp_file.exists()
    else:
        assert temp_file.exists()


def test_finalize_within_scope(tmp_path: pathlib.Path):
    temp_file = tmp_path / "test_finalize_within_scope.hatchet"
    session_id0 = proton.start(str(temp_file.with_suffix("")))
    with proton.scope("test0"):
        assert proton.context.depth(session_id0) == 1
        proton.finalize()
    assert temp_file.exists()
    temp_file1 = tmp_path / "test_finalize_within_scope1.hatchet"
    session_id1 = proton.start(str(temp_file1.with_suffix("")))
    depth = proton.context.depth(session_id1)
    assert depth == 0
    proton.finalize()


def test_data_api(tmp_path: pathlib.Path):
    temp_file = tmp_path / "test_data_api.hatchet"
    session_id = proton.start(str(temp_file.with_suffix("")))
    proton.enter_scope("test0")
    proton.exit_scope()
    proton.deactivate(session_id)
    json_data = proton.data.get(session_id)
    assert json_data is not None
    msgpack_data = proton.data.get_msgpack(session_id)
    assert isinstance(msgpack_data, bytes)
    is_complete = proton.data.is_phase_complete(session_id, 0)
    assert is_complete is False
    next_phase = proton.data.advance_phase(session_id)
    assert next_phase == 1
    is_complete = proton.data.is_phase_complete(session_id, 1)
    assert is_complete is False

    # Even if a phase has no GPU activity records, flushing should still mark it
    # as flushed.
    proton.activate(session_id)
    next_phase = proton.data.advance_phase(session_id)
    assert next_phase == 2
    proton.deactivate(session_id, flushing=True)
    assert proton.data.is_phase_complete(session_id, 1) is True
    assert proton.data.is_phase_complete(session_id, 2) is False

    # Test clear and clear_up_to_phase
    proton.data.clear(session_id, phase=0)
    proton.data.clear(session_id, phase=2, clear_up_to_phase=True)

    proton.finalize()



"""
Reproducibility tests for Proton.
Each test should invoke one or more GPU kernels and check the validity of their profiling results.
"""

import torch
import triton
import triton.profiler as proton
import json
import pytest
from typing import NamedTuple
import pathlib
import threading

import triton.language as tl
from triton.profiler.hooks.launch import COMPUTE_METADATA_SCOPE_NAME
import triton.profiler.hooks.launch as proton_launch
import triton.profiler.viewer as viewer
from triton._internal_testing import is_hip, is_cuda, is_blackwell


@pytest.mark.parametrize("context", ["shadow", "python"])
def test_torch(context, tmp_path: pathlib.Path, device: str):
    temp_file = tmp_path / "test_torch.hatchet"
    proton.start(str(temp_file.with_suffix("")), context=context)
    proton.enter_scope("test")
    torch.ones((2, 2), device=device)
    proton.exit_scope()
    proton.finalize()
    with temp_file.open() as f:
        data = json.load(f)
    if context == "shadow":
        assert len(data[0]["children"]) == 1
        assert data[0]["children"][0]["frame"]["name"] == "test"
        assert data[0]["children"][0]["children"][0]["metrics"]["time (ns)"] > 0
    elif context == "python":
        assert len(data[0]["children"]) == 1
        # bfs search until find the "elementwise_kernel" and then check its children
        queue = [data[0]]
        import re
        while len(queue) > 0:
            parent_frame = queue.pop(0)
            for child in parent_frame["children"]:
                if "elementwise_kernel" in child["frame"]["name"]:
                    assert len(child["children"]) == 0
                    # check the regex of the parent name matches
                    # file_name:line_number@function_name
                    regex = r".+:\d+@.+"
                    assert re.match(regex, parent_frame["frame"]["name"])
                    return
                queue.append(child)


def test_triton(tmp_path: pathlib.Path, device: str):

    @triton.jit
    def foo(x, y):
        tl.store(y, tl.load(x))

    x = torch.tensor([2], device=device)
    y = torch.zeros_like(x)
    temp_file = tmp_path / "test_triton.hatchet"
    proton.start(str(temp_file.with_suffix("")))
    with proton.scope("test0"):
        with proton.scope("test1"):
            foo[(1, )](x, y)
    with proton.scope("test2"):
        foo[(1, )](x, y)
    proton.finalize()
    with temp_file.open() as f:
        data = json.load(f)
    assert len(data[0]["children"]) == 2
    assert data[0]["children"][0]["frame"]["name"] == "test0"
    assert len(data[0]["children"][0]["children"]) == 1
    assert data[0]["children"][0]["children"][0]["frame"]["name"] == "test1"
    assert data[0]["children"][1]["frame"]["name"] == "test2"


@pytest.mark.skipif(not is_cuda(), reason="HIP backend does not reliably attribute cudagraph replay launches to scopes")
def test_cudagraph(tmp_path: pathlib.Path, device: str):
    stream = torch.cuda.Stream()
    torch.cuda.set_stream(stream)

    def metadata_fn(grid: tuple, metadata: NamedTuple, args: dict):
        return {"name": "foo_test"}

    @triton.jit(launch_metadata=metadata_fn)
    def foo(x, y, z):
        tl.store(z, tl.load(y) + tl.load(x))

    def fn():
        a = torch.ones((2, 2), device=device)
        b = torch.ones((2, 2), device=device)
        c = a + b
        foo[(1, )](a, b, c)

    temp_file = tmp_path / "test_cudagraph.hatchet"
    proton.start(str(temp_file.with_suffix("")), context="shadow")

    # warmup
    # four kernels
    fn()

    # no kernels
    g = torch.cuda.CUDAGraph()
    with torch.cuda.graph(g):
        for i in range(10):
            with proton.scope(f"iter_{i}"):
                fn()

    with proton.scope("test0"):
        g.replay()

    with proton.scope("test1"):
        g.replay()

    g.reset()
    proton.finalize()

    with temp_file.open() as f:
        data = json.load(f)
    # CUDA/HIP graph may also invoke additional kernels to reset outputs
    # {torch.ones, add, foo, test}
    assert len(data[0]["children"]) >= 4
    # find the test frame
    test0_frame = None
    test1_frame = None
    for child in data[0]["children"]:
        if child["frame"]["name"] == "test0":
            test0_frame = child
        if child["frame"]["name"] == "test1":
            test1_frame = child
    assert test0_frame is not None
    assert test1_frame is not None
    # {torch.ones, add, foo}
    if is_hip():
        assert len(test0_frame["children"]) >= 2
        assert test0_frame["children"][0]["metrics"]["time (ns)"] > 0
    else:
        # cuda backend supports "<captured_at>" annotation
        for test_frame in [test0_frame, test1_frame]:
            child = test_frame["children"][0]
            assert child["frame"]["name"] == "<captured_at>"
            # 0...9 iterations
            assert len(child["children"]) == 10
            # check all iterations
            for i in range(10):
                assert child["children"][i]["frame"]["name"] == f"iter_{i}"
                assert child["children"][i]["children"][0]["metrics"]["time (ns)"] > 0


@pytest.mark.skipif(not is_cuda(), reason="Only CUDA backend supports cudagraph replay")
def test_cudagraph_not_captured_by_profiler(tmp_path: pathlib.Path, capfd, device: str):
    stream = torch.cuda.Stream()
    torch.cuda.set_stream(stream)

    @triton.jit
    def foo(x, y, z):
        tl.store(z, tl.load(y) + tl.load(x))

    def fn():
        a = torch.ones((2, 2), device=device)
        b = torch.ones((2, 2), device=device)
        c = a + b
        foo[(1, )](a, b, c)

    # Build/capture graph before profiler starts.
    fn()
    g = torch.cuda.CUDAGraph()
    with torch.cuda.graph(g):
        fn()

    temp_file = tmp_path / "test_cudagraph_not_captured_by_profiler.hatchet"
    proton.start(str(temp_file.with_suffix("")), context="shadow")
    with proton.scope("replay0"):
        g.replay()
    with proton.scope("replay1"):
        g.replay()
    proton.finalize()

    captured = capfd.readouterr()
    assert captured.err.count("Cannot find graph for graphExecId:") == 1
    assert "start profiling before the graph is created" in captured.err

    with temp_file.open() as f:
        data = json.load(f)
    replay0_frame = None
    replay1_frame = None
    for child in data[0]["children"]:
        if child["frame"]["name"] == "replay0":
            replay0_frame = child
        elif child["frame"]["name"] == "replay1":
            replay1_frame = child
    assert replay0_frame is not None
    assert replay1_frame is not None
    assert len(replay0_frame["children"]) == 3
    assert len(replay1_frame["children"]) == 3

    def has_positive_time_metric(node):
        if node["metrics"].get("time (ns)", 0) > 0:
            return True
        return any(has_positive_time_metric(child) for child in node["children"])

    assert has_positive_time_metric(replay0_frame)
    assert has_positive_time_metric(replay1_frame)


@pytest.mark.skipif(not is_cuda(), reason="Only CUDA backend supports cudagraph deactivation")
def test_cudagraph_deactivate(tmp_path, device: str):
    stream = torch.cuda.Stream()
    torch.cuda.set_stream(stream)

    @triton.jit
    def foo(x, y, z):
        tl.store(z, tl.load(y) + tl.load(x))

    def fn(session):
        with proton.scope("scope_a"):
            a = torch.ones((2, 2), device=device)
        proton.deactivate(session)
        with proton.scope("scope_b"):
            b = torch.ones((2, 2), device=device)
        proton.activate(session)
        with proton.scope("scope_c"):
            c = a + b
        foo[(1, )](a, b, c)

    temp_file = tmp_path / "test_cudagraph_deactivate.hatchet"
    session = proton.start(str(temp_file.with_suffix("")), context="shadow", hook="triton")

    # warmup
    fn(session)

    # no kernels
    g = torch.cuda.CUDAGraph()
    with torch.cuda.graph(g):
        for i in range(10):
            with proton.scope(f"iter_{i}"):
                fn(session)

    with proton.scope("test0"):
        g.replay()

    proton.finalize()

    with temp_file.open() as f:
        data = json.load(f)

    # scope a and c should be recorded, b should be skipped
    children = data[0]["children"]
    test0_frame = None
    for child in children:
        if child["frame"]["name"] == "test0":
            test0_frame = child
            break
    assert test0_frame is not None
    iter_frame = test0_frame["children"][0]["children"][0]
    scope_a_frame = None
    scope_b_frame = None
    scope_c_frame = None
    for child in iter_frame["children"]:
        if child["frame"]["name"] == "scope_a":
            scope_a_frame = child
        if child["frame"]["name"] == "scope_b":
            scope_b_frame = child
        if child["frame"]["name"] == "scope_c":
            scope_c_frame = child
    assert scope_a_frame is not None
    assert scope_b_frame is None
    assert scope_c_frame is not None


def test_metrics(tmp_path: pathlib.Path, device: str):

    @triton.jit
    def foo(x, y):
        tl.store(y, tl.load(x))

    x = torch.tensor([2], device=device)
    y = torch.zeros_like(x)
    temp_file = tmp_path / "test_metrics.hatchet"
    proton.start(str(temp_file.with_suffix("")))
    with proton.scope("test0", {"foo": 1.0, "bar": [1, 2, 3], "baz": [1.0, 2.0, 3.0]}):
        foo[(1, )](x, y)
    proton.finalize()
    with temp_file.open() as f:
        data = json.load(f)
    assert len(data[0]["children"]) == 1
    assert data[0]["children"][0]["frame"]["name"] == "test0"
    assert data[0]["children"][0]["metrics"]["foo"] == 1.0
    assert data[0]["children"][0]["metrics"]["bar"] == [1, 2, 3]


def test_scope_backward(tmp_path: pathlib.Path, device: str):
    temp_file = tmp_path / "test_scope_backward.hatchet"
    proton.start(str(temp_file.with_suffix("")))
    with proton.scope("ones1"):
        a = torch.ones((100, 100), device=device, requires_grad=True)
    with proton.scope("plus"):
        a2 = a * a * a
    with proton.scope("ones2"):
        loss = torch.ones_like(a2)

    # Backward triggers two kernels in a single scope
    with proton.scope("backward"):
        a2.backward(loss)
    proton.finalize()
    with temp_file.open() as f:
        data = json.load(f)
    assert len(data[0]["children"]) == 4


def test_cpu_timed_scope(tmp_path: pathlib.Path, device: str):
    temp_file = tmp_path / "test_cpu_timed_scope.hatchet"
    proton.start(str(temp_file.with_suffix("")))
    with proton.cpu_timed_scope("test0"):
        with proton.cpu_timed_scope("test1"):
            torch.ones((100, 100), device=device)
    proton.finalize()
    with temp_file.open() as f:
        data = json.load(f)
    assert len(data[0]["children"]) == 1
    test0_frame = data[0]["children"][0]
    assert test0_frame["metrics"]["cpu_time (ns)"] > 0
    test1_frame = test0_frame["children"][0]
    assert test1_frame["metrics"]["cpu_time (ns)"] > 0


def test_get_data(tmp_path: pathlib.Path, device: str):
    temp_file = tmp_path / "test_tree_json.hatchet"
    session = proton.start(str(temp_file.with_suffix("")), context="shadow")

    @triton.jit
    def foo(x, y, size: tl.constexpr):
        offs = tl.arange(0, size)
        tl.store(y + offs, tl.load(x + offs))

    with proton.scope("test"):
        x = torch.ones((2, 2), device=device)
        foo[(1, )](x, x, 4)
        foo[(1, )](x, x, 4)

    proton.deactivate(session, flushing=True)

    database = proton.data.get(session)
    gf, _, _, _ = viewer.get_raw_metrics(database)
    foo_frame = gf.filter("MATCH ('*', c) WHERE c.'name' =~ '.*foo.*' AND c IS LEAF").dataframe
    ones_frame = gf.filter("MATCH ('*', c) WHERE c.'name' =~ '.*elementwise.*' AND c IS LEAF").dataframe

    assert len(foo_frame) == 1
    assert int(foo_frame["count"].values[0]) == 2
    assert len(ones_frame) == 1
    assert int(ones_frame["count"].values[0]) == 1

    import msgpack
    msgpack_data = proton.data.get_msgpack(session)
    database_unpacked = msgpack.loads(msgpack_data, raw=False, strict_map_key=False)
    assert database == database_unpacked

    proton.finalize()


def test_clear_data(tmp_path: pathlib.Path, device: str):
    temp_file = tmp_path / "test_clear_data.hatchet"
    session = proton.start(str(temp_file.with_suffix("")), context="shadow")

    with proton.scope("test0"):
        x = torch.ones((2, 2), device=device)
        x + x  # type: ignore

    proton.deactivate(session, flushing=True)
    proton.data.clear(session)
    try:
        database = proton.data.get(session)
    except RuntimeError as e:
        assert "has no data" in str(e)

    proton.activate(session)
    with proton.scope("test1"):
        x * x  # type: ignore
    proton.deactivate(session, flushing=True)
    database = proton.data.get(session)

    proton.finalize()
    assert len(database[0]["children"]) == 1
    assert database[0]["children"][0]["frame"]["name"] == "test1"
    kernel_frame = database[0]["children"][0]["children"][0]
    assert "elementwise" in kernel_frame["frame"]["name"]


def test_clear_data_up_to_phase(tmp_path: pathlib.Path, device: str):
    temp_file = tmp_path / "test_clear_data_up_to_phase.hatchet"
    session = proton.start(str(temp_file.with_suffix("")), context="shadow")

    with proton.scope("phase0"):
        x = torch.ones((2, 2), device=device)
        x + x  # type: ignore

    phase1 = proton.data.advance_phase(session)
    with proton.scope("phase1"):
        x = torch.ones((2, 2), device=device)
        x + x  # type: ignore

    proton.deactivate(session, flushing=True)

    # Clear a range of phases.
    proton.data.clear(session, phase=phase1, clear_up_to_phase=True)
    database = proton.data.get(session, phase=phase1)
    assert len(database[0]["children"]) == 0

    proton.finalize()


def test_data_is_phase_complete(tmp_path: pathlib.Path, device: str):
    temp_path = tmp_path / "test_data_is_phase_complete.hatchet"
    session = proton.start(str(temp_path.with_suffix("")), context="shadow")

    def fn():
        with proton.scope("test0"):
            x = torch.ones((2, 2), device=device)
            x + x  # type: ignore

    fn()
    assert not proton.data.is_phase_complete(session, 0)

    proton.deactivate(session)
    # likely the GPU has not completed the data yet
    assert not proton.data.is_phase_complete(session, 0)

    proton.activate(session)
    phase = proton.data.advance_phase(session)
    fn()
    proton.deactivate(session, flushing=True)
    # session 0 is a previous phase but we have called deactivate with flushing
    assert proton.data.is_phase_complete(session, 0)
    # phase 1 is the current phase so cannot be a completed phase
    assert not proton.data.is_phase_complete(session, phase)
    proton.data.advance_phase(session)
    # phase 0 should remain completed after advancing phases
    assert proton.data.is_phase_complete(session, phase - 1)

    proton.finalize()


def test_hook_launch(tmp_path: pathlib.Path, device: str):

    def metadata_fn(grid: tuple, metadata: NamedTuple, args: dict):
        # get arg's element size
        element_size = args["x"].element_size()  # non-const
        size = args["size"]  # const
        key = "flops" + str(element_size * 8)
        num_ctas = metadata.num_ctas
        # Return an extra metric key beyond the historical flops/bytes allowlist.
        return {"name": f"foo_test_{num_ctas}ctas_{size}elems", key: 1.0, "extra_metric": 7.0}

    @triton.jit(launch_metadata=metadata_fn)
    def foo(x, size: tl.constexpr, y):
        offs = tl.arange(0, size)
        tl.store(y + offs, tl.load(x + offs))

    x = torch.tensor([2], device=device, dtype=torch.float32)
    y = torch.zeros_like(x)
    temp_file = tmp_path / "test_hook_triton.hatchet"
    proton.start(str(temp_file.with_suffix("")), hook="triton")
    with proton.scope("test0"):
        foo[(1, )](x, 1, y, num_warps=4)
    proton.finalize()
    with temp_file.open() as f:
        data = json.load(f)
    assert len(data[0]["children"]) == 1
    assert data[0]["children"][0]["frame"]["name"] == "test0"
    assert data[0]["children"][0]["children"][0]["frame"]["name"] == "foo_test_1ctas_1elems"
    assert data[0]["children"][0]["children"][0]["metrics"]["flops32"] == 1.0
    assert data[0]["children"][0]["children"][0]["metrics"]["extra_metric"] == 7.0
    assert data[0]["children"][0]["children"][0]["metrics"]["time (ns)"] > 0


def test_hook_launch_filter(tmp_path: pathlib.Path, device: str):

    foo_metadata_invoked = False
    bar_metadata_invoked = False

    def foo_metadata_fn(grid: tuple, metadata: NamedTuple, args: dict):
        nonlocal foo_metadata_invoked
        foo_metadata_invoked = True
        return {"name": "foo_meta", "flops": 1.0}

    def bar_metadata_fn(grid: tuple, metadata: NamedTuple, args: dict):
        nonlocal bar_metadata_invoked
        bar_metadata_invoked = True
        return {"name": "bar_meta", "flops": 2.0}

    @triton.jit(launch_metadata=foo_metadata_fn)
    def foo(x, size: tl.constexpr, y):
        offs = tl.arange(0, size)
        tl.store(y + offs, tl.load(x + offs))

    @triton.jit(launch_metadata=bar_metadata_fn)
    def bar(x, size: tl.constexpr, y):
        offs = tl.arange(0, size)
        tl.store(y + offs, tl.load(x + offs))

    x = torch.tensor([2], device=device, dtype=torch.float32)
    y = torch.zeros_like(x)
    temp_file = tmp_path / "test_hook_triton_filter.hatchet"

    # Only allow kernels whose compiled name matches "foo" (via prefix regex).
    launch_hook = proton_launch.LaunchHook()
    launch_hook.configure(include=".*foo")
    proton.start(str(temp_file.with_suffix("")), hook=launch_hook)
    with proton.scope("test0"):
        foo[(1, )](x, 1, y, num_warps=4)
        bar[(1, )](x, 1, y, num_warps=4)
    proton.finalize()
    # Reset singleton hook state to avoid leaking filter settings across tests.
    launch_hook.configure(include=None, exclude=None)

    assert foo_metadata_invoked is True
    assert bar_metadata_invoked is False

    with temp_file.open() as f:
        data = json.load(f)

    # Ensure the "foo_meta" override exists and "bar_meta" does not.
    all_names = set()
    queue = [data[0]]
    while queue:
        node = queue.pop()
        if "frame" in node and "name" in node["frame"]:
            all_names.add(node["frame"]["name"])
        queue.extend(node.get("children", []))

    assert "foo_meta" in all_names
    assert "bar_meta" not in all_names


@pytest.mark.parametrize("context", ["shadow", "python"])
def test_hook_launch_context(tmp_path: pathlib.Path, context: str, device: str):

    def metadata_fn(grid: tuple, metadata: NamedTuple, args: dict):
        x = args["x"]
        # A gpu kernel, but it should be under the metadata state
        return {"name": "foo_test", "bytes": x.sum().item()}

    @triton.jit(launch_metadata=metadata_fn)
    def foo(x, size: tl.constexpr, y):
        offs = tl.arange(0, size)
        tl.store(y + offs, tl.load(x + offs))

    x = torch.tensor([2], device=device, dtype=torch.float32)
    y = torch.zeros_like(x)
    temp_file = tmp_path / "test_hook.hatchet"
    proton.start(str(temp_file.with_suffix("")), hook="triton", context=context)
    with proton.scope("test0"):
        foo[(1, )](x, 1, y, num_warps=4)
    proton.finalize()
    with temp_file.open() as f:
        data = json.load(f)
    # bfs search until find the reduce kernel and then check its parent
    queue = [data[0]]
    while len(queue) > 0:
        parent_frame = queue.pop(0)
        for child in parent_frame["children"]:
            if "reduce" in child["frame"]["name"]:
                assert parent_frame["frame"]["name"] == COMPUTE_METADATA_SCOPE_NAME
                return
            queue.append(child)


def test_hook_with_third_party(tmp_path: pathlib.Path, device: str):
    third_party_hook_invoked = False

    def third_party_hook(metadata) -> None:
        nonlocal third_party_hook_invoked
        third_party_hook_invoked = True

    triton.knobs.runtime.launch_enter_hook.add(third_party_hook)

    proton_hook_invoked = False

    def metadata_fn(grid: tuple, metadata: NamedTuple, args: dict):
        nonlocal proton_hook_invoked
        proton_hook_invoked = True
        return {"name": "foo_test"}

    @triton.jit(launch_metadata=metadata_fn)
    def foo(x, size: tl.constexpr, y):
        offs = tl.arange(0, size)
        tl.store(y + offs, tl.load(x + offs))

    x = torch.tensor([2], device=device, dtype=torch.float32)
    y = torch.zeros_like(x)
    temp_file = tmp_path / "test_hook_with_third_party.hatchet"
    proton.start(str(temp_file.with_suffix("")), hook="triton")
    foo[(1, )](x, 1, y, num_warps=4)
    proton.finalize()
    triton.knobs.runtime.launch_enter_hook.remove(third_party_hook)
    with temp_file.open() as f:
        data = json.load(f)
    assert len(data[0]["children"]) == 1
    assert data[0]["children"][0]["frame"]["name"] == "foo_test"
    assert data[0]["children"][0]["metrics"]["time (ns)"] > 0


def test_hook_multiple_threads(tmp_path: pathlib.Path, device: str):

    def metadata_fn_foo(grid: tuple, metadata: NamedTuple, args: dict):
        return {"name": "foo_test"}

    @triton.jit(launch_metadata=metadata_fn_foo)
    def foo(x, size: tl.constexpr, y):
        offs = tl.arange(0, size)
        tl.store(y + offs, tl.load(x + offs))

    def metadata_fn_bar(grid: tuple, metadata: NamedTuple, args: dict):
        return {"name": "bar_test"}

    @triton.jit(launch_metadata=metadata_fn_bar)
    def bar(x, size: tl.constexpr, y):
        offs = tl.arange(0, size)
        tl.store(y + offs, tl.load(x + offs))

    x_foo = torch.tensor([2], device=device, dtype=torch.float32)
    y_foo = torch.zeros_like(x_foo)
    x_bar = torch.tensor([2], device=device, dtype=torch.float32)
    y_bar = torch.zeros_like(x_bar)

    temp_file = tmp_path / "test_hook.hatchet"
    proton.start(str(temp_file.with_suffix("")), hook="triton")

    all_ids = set()

    # start multiple threads
    def invoke_foo():
        for _ in range(100):
            foo[(1, )](x_foo, 1, y_foo, num_warps=4)
            all_ids.add(proton_launch.id.get())

    def invoke_bar():
        for _ in range(100):
            bar[(1, )](x_bar, 1, y_bar, num_warps=4)
            all_ids.add(proton_launch.id.get())

    thread_foo = threading.Thread(target=invoke_foo)
    thread_bar = threading.Thread(target=invoke_bar)
    thread_foo.start()
    thread_bar.start()
    thread_foo.join()
    thread_bar.join()

    proton.finalize()
    assert len(all_ids) == 200

    with temp_file.open() as f:
        data = json.load(f)
    root = data[0]["children"]
    assert "foo_test" in root[0]["frame"]["name"] or root[1]["frame"]["name"]
    assert "bar_test" in root[0]["frame"]["name"] or root[1]["frame"]["name"]
    assert root[0]["metrics"]["count"] == 100
    assert root[1]["metrics"]["count"] == 100


def test_pcsampling(tmp_path: pathlib.Path, device: str):
    if not is_cuda():
        pytest.skip("Only CUDA backend supports pc sampling")

    import os

    if os.environ.get("PROTON_SKIP_PC_SAMPLING_TEST", "0") == "1":
        pytest.skip("PC sampling test is disabled")

    @triton.jit
    def foo(x, y, size: tl.constexpr):
        offs = tl.arange(0, size)
        for _ in range(1000):
            tl.store(y + offs, tl.load(x + offs))

    temp_file = tmp_path / "test_pcsampling.hatchet"
    proton.start(str(temp_file.with_suffix("")), hook="triton", backend="cupti", mode="pcsampling")
    with proton.scope("init"):
        x = torch.ones((1024, ), device=device, dtype=torch.float32)
        y = torch.zeros_like(x)
    with proton.scope("test"):
        foo[(1, )](x, y, x.size()[0], num_warps=4)
    proton.finalize()
    with temp_file.open() as f:
        data = json.load(f)
    init_frame = data[0]["children"][0]
    test_frame = data[0]["children"][1]
    # With line mapping
    assert "foo" in test_frame["children"][0]["frame"]["name"]
    assert test_frame["children"][0]["children"][0]["metrics"]["num_samples"] > 0
    assert "@" in test_frame["children"][0]["children"][0]["frame"]["name"]
    # Without line mapping
    assert "elementwise" in init_frame["children"][0]["frame"]["name"]
    assert init_frame["children"][0]["metrics"]["num_samples"] > 0


def test_deactivate(tmp_path: pathlib.Path, device: str):
    temp_file = tmp_path / "test_deactivate.hatchet"
    session_id = proton.start(str(temp_file.with_suffix("")), hook="triton")
    proton.deactivate(session_id)
    torch.randn((10, 10), device=device)
    proton.activate(session_id)
    torch.zeros((10, 10), device=device)
    proton.deactivate(session_id)
    proton.finalize()
    with temp_file.open() as f:
        data = json.load(f)
    # Root shouldn't have device id
    assert "device_id" not in data[0]["metrics"]
    assert len(data[0]["children"]) == 1
    assert "device_id" in data[0]["children"][0]["metrics"]


def test_multiple_sessions(tmp_path: pathlib.Path, device: str):
    temp_file0 = tmp_path / "test_multiple_sessions0.hatchet"
    temp_file1 = tmp_path / "test_multiple_sessions1.hatchet"
    session_id0 = proton.start(str(temp_file0.with_suffix("")))
    session_id1 = proton.start(str(temp_file1.with_suffix("")))
    with proton.scope("scope0"):
        torch.randn((10, 10), device=device)
        torch.randn((10, 10), device=device)
    proton.deactivate(session_id0)
    proton.finalize(session_id0)
    with proton.scope("scope1"):
        torch.randn((10, 10), device=device)
    proton.finalize(session_id1)
    # kernel has been invoked twice in session 0 and three times in session 1
    with temp_file0.open() as f:
        data = json.load(f)
    assert data[0]["children"][0]["frame"]["name"] == "scope0"
    assert int(data[0]["children"][0]["children"][0]["metrics"]["count"]) == 2
    with temp_file1.open() as f:
        data = json.load(f)
    scope0_count = int(data[0]["children"][0]["children"][0]["metrics"]["count"])
    scope1_count = int(data[0]["children"][1]["children"][0]["metrics"]["count"])
    assert scope0_count + scope1_count == 3


@pytest.mark.skipif(not is_cuda(), reason="Only CUDA backend supports metrics profiling in cudagraphs")
def test_multiple_sessions_cudagraph_metric_kernels(tmp_path: pathlib.Path, device: str):
    stream = torch.cuda.Stream()
    torch.cuda.set_stream(stream)

    foo_iters = 3
    bar_iters = 2

    def foo_metadata_fn(grid: tuple, metadata: NamedTuple, args: dict):
        x = args["x"]
        # Tensor custom metric in graph capture mode launches metric kernels.
        return {"name": "foo_with_metric", "sum_metric": x.sum()}

    def bar_metadata_fn(grid: tuple, metadata: NamedTuple, args: dict):
        # Name-only metadata (no custom metric).
        return {"name": "bar_without_metric"}

    @triton.jit(launch_metadata=foo_metadata_fn)
    def foo(x, y, z):
        tl.store(z, tl.load(y) + tl.load(x))

    @triton.jit(launch_metadata=bar_metadata_fn)
    def bar(x, y, z):
        tl.store(z, tl.load(y) - tl.load(x))

    x = torch.ones((2, 2), device=device)
    y = torch.ones((2, 2), device=device)
    z = torch.empty_like(x)

    # Compile kernels before profiling starts to reduce unrelated profile noise.
    foo[(1, )](x, y, z)
    bar[(1, )](x, y, z)

    temp_file0 = tmp_path / "test_multiple_sessions_cudagraph_metric_kernels0.hatchet"
    temp_file1 = tmp_path / "test_multiple_sessions_cudagraph_metric_kernels1.hatchet"
    session_id0 = proton.start(str(temp_file0.with_suffix("")), context="shadow", hook="triton")
    session_id1 = proton.start(str(temp_file1.with_suffix("")), context="shadow", hook="triton")

    proton.deactivate(session_id1)

    graph_foo = torch.cuda.CUDAGraph()
    with torch.cuda.graph(graph_foo):
        for _ in range(foo_iters):
            foo[(1, )](x, y, z)
    with proton.scope("session0_replay"):
        graph_foo.replay()

    proton.deactivate(session_id0)
    proton.activate(session_id1)

    graph_bar = torch.cuda.CUDAGraph()
    with torch.cuda.graph(graph_bar):
        for _ in range(bar_iters):
            bar[(1, )](x, y, z)
    with proton.scope("session1_replay"):
        graph_bar.replay()

    proton.finalize(session_id0)
    proton.finalize(session_id1)

    def get_frame_by_name(node, name: str):
        queue = [node]
        while queue:
            cur = queue.pop(0)
            if cur["frame"]["name"] == name:
                return cur
            queue.extend(cur["children"])
        return None

    def get_all_names(node):
        names = set()
        queue = [node]
        while queue:
            cur = queue.pop(0)
            names.add(cur["frame"]["name"])
            queue.extend(cur["children"])
        return names

    with temp_file0.open() as f:
        data0 = json.load(f)
    with temp_file1.open() as f:
        data1 = json.load(f)

    session0_replay_frame = get_frame_by_name(data0[0], "session0_replay")
    session1_replay_frame = get_frame_by_name(data1[0], "session1_replay")
    assert session0_replay_frame is not None
    assert session1_replay_frame is not None

    capture0 = session0_replay_frame["children"][0]
    capture1 = session1_replay_frame["children"][0]

    foo_frame0 = get_frame_by_name(capture0, "foo_with_metric")
    bar_frame0 = get_frame_by_name(capture0, "bar_without_metric")
    foo_frame1 = get_frame_by_name(capture1, "foo_with_metric")
    bar_frame1 = get_frame_by_name(capture1, "bar_without_metric")

    assert foo_frame0 is not None
    assert bar_frame0 is None
    assert foo_frame1 is None
    assert bar_frame1 is not None

    assert foo_frame0["metrics"]["sum_metric"] == float(foo_iters * x.numel())
    assert int(foo_frame0["metrics"]["count"]) == foo_iters
    assert "sum_metric" not in bar_frame1["metrics"]
    assert int(bar_frame1["metrics"]["count"]) == bar_iters


def test_trace(tmp_path: pathlib.Path, device: str):
    temp_file = tmp_path / "test_trace.chrome_trace"
    proton.start(str(temp_file.with_suffix("")), data="trace")

    @triton.jit
    def foo(x, y, size: tl.constexpr):
        offs = tl.arange(0, size)
        tl.store(y + offs, tl.load(x + offs))

    with proton.scope("init"):
        x = torch.ones((1024, ), device=device, dtype=torch.float32)
        y = torch.zeros_like(x)

    with proton.scope("test"):
        foo[(1, )](x, y, x.size()[0], num_warps=4)

    proton.finalize()

    with temp_file.open() as f:
        data = json.load(f)
        trace_events = data["traceEvents"]
        assert len(trace_events) == 3
        assert trace_events[-1]["name"] == "foo"
        assert trace_events[-1]["args"]["call_stack"] == ["ROOT", "test", "foo"]


def test_scope_multiple_threads(tmp_path: pathlib.Path, device: str):
    temp_file = tmp_path / "test_scope_threads.hatchet"
    proton.start(str(temp_file.with_suffix("")))

    N = 50
    thread_names = ["threadA", "threadB"]

    def worker(prefix: str):
        for i in range(N):
            name = f"{prefix}_{i}"
            proton.enter_scope(name)
            torch.ones((1, ), device=device)
            proton.exit_scope()

    threads = [threading.Thread(target=worker, args=(tname, )) for tname in thread_names]
    for t in threads:
        t.start()
    for t in threads:
        t.join()

    proton.finalize()

    with temp_file.open() as f:
        data = json.load(f)

    children = data[0]["children"]
    assert len(children) == N * len(thread_names)
    names = {c["frame"]["name"] for c in children}
    expected = {f"{t}_{i}" for t in thread_names for i in range(N)}
    assert names == expected


@pytest.mark.skipif(not is_cuda() and not is_hip(), reason="Only CUDA/HIP backend supports NVTX profiling")
@pytest.mark.parametrize("enable_nvtx", [None, True, False])
def test_nvtx_range_push_pop(enable_nvtx, fresh_knobs, tmp_path: pathlib.Path, device: str):
    if enable_nvtx is not None:
        fresh_knobs.proton.enable_nvtx = enable_nvtx
    temp_file = tmp_path / "test_nvtx_range_push_pop.hatchet"
    proton.start(str(temp_file.with_suffix("")))

    with proton.scope("proton_scope"):
        torch.cuda.nvtx.range_push("nvtx_range0")
        torch.cuda.nvtx.range_push("nvtx_range1")
        torch.ones((1, ), device=device)
        torch.cuda.nvtx.range_pop()
        torch.cuda.nvtx.range_pop()

    proton.finalize()

    with temp_file.open() as f:
        data = json.load(f)

    children = data[0]["children"]
    assert len(children) == 1
    proton_scope = children[0]
    assert proton_scope["frame"]["name"] == "proton_scope"
    assert len(proton_scope["children"]) == 1
    if enable_nvtx or enable_nvtx is None:
        nvtx_range0 = proton_scope["children"][0]
        assert nvtx_range0["frame"]["name"] == "nvtx_range0"
        assert len(nvtx_range0["children"]) == 1
        nvtx_range1 = nvtx_range0["children"][0]
        assert nvtx_range1["frame"]["name"] == "nvtx_range1"
        assert len(nvtx_range1["children"]) == 1
        kernel = nvtx_range1["children"][0]
    else:
        kernel = proton_scope["children"][0]
    assert "elementwise" in kernel["frame"]["name"]
    assert kernel["metrics"]["count"] == 1


def test_tensor_metrics_scope(tmp_path: pathlib.Path, device: str):
    temp_file = tmp_path / "test_tensor_metrics_scope.hatchet"
    proton.start(str(temp_file.with_suffix("")))

    x = torch.ones((10, 10), device=device, dtype=torch.float32)
    x_mean = x.mean()
    x_std = x.std()
    with proton.scope("test", metrics={"x_mean": x_mean, "x_std": x_std}):
        torch.randn((10, 10), device=device)
        torch.zeros_like(x)

    proton.finalize()

    with temp_file.open() as f:
        data = json.load(f)

    children = data[0]["children"]
    assert len(children) == 4
    # get the test frame
    test_frame = None
    for child in children:
        if child["frame"]["name"] == "test":
            test_frame = child
            break
    assert test_frame is not None
    assert test_frame["metrics"]["x_mean"] == 1.0
    assert test_frame["metrics"]["x_std"] == 0.0


def test_tensor_metrics_hook(tmp_path: pathlib.Path, device: str):
    temp_file = tmp_path / "test_tensor_metrics_hook.hatchet"

    def metadata_fn(grid: tuple, metadata: NamedTuple, args: dict):
        metric_value = torch.tensor(8.0, device=device)
        return {"name": "foo_test", "flops": metric_value}

    @triton.jit(launch_metadata=metadata_fn)
    def foo(x, size: tl.constexpr, y):
        offs = tl.arange(0, size)
        tl.store(y + offs, tl.load(x + offs))

    x = torch.ones((8, ), device=device, dtype=torch.float32)
    y = torch.zeros_like(x)

    proton.start(str(temp_file.with_suffix("")), hook="triton")
    foo[(1, )](x, x.numel(), y, num_warps=4)
    proton.finalize()

    with temp_file.open() as f:
        data = json.load(f)

    children = data[0]["children"]
    # metadata scope + foo_test
    assert len(children) == 2
    foo_test_frame = None
    for child in children:
        if child["frame"]["name"] == "foo_test":
            foo_test_frame = child
            break
    assert foo_test_frame is not None
    assert foo_test_frame["metrics"]["flops"] == 8.0


@pytest.mark.skipif(not is_cuda(), reason="Only CUDA backend supports metrics profiling in cudagraphs")
def test_tensor_metrics_cudagraph(tmp_path: pathlib.Path, device: str):
    stream = torch.cuda.Stream()
    torch.cuda.set_stream(stream)

    def metadata_fn(grid: tuple, metadata: NamedTuple, args: dict):
        x = args["x"]
        x_sum = x.sum()
        return {"name": "foo_test", "bytes": x.numel() * x.element_size(), "flops": x_sum}

    @triton.jit(launch_metadata=metadata_fn)
    def foo(x, y, z):
        tl.store(z, tl.load(y) + tl.load(x))

    def fn():
        with proton.scope("scope_a", metrics={"bytes": 4 * 4}):
            a = torch.ones((2, 2), device=device)
        with proton.metadata_state():
            a_sum = a.sum()
        with proton.scope("scope_b", metrics={"sum": a_sum}):
            b = torch.ones((2, 2), device=device)
        c = a + b
        foo[(1, )](a, b, c)
        with proton.metadata_state():
            d = torch.arange(4, device="cuda")
        with proton.scope("scope_d", metrics={"vec": d}):
            e = d * 2  # noqa: F841

    temp_file = tmp_path / "test_tensor_metrics_cudagraph.hatchet"
    proton.start(str(temp_file.with_suffix("")), context="shadow", hook="triton")

    # warmup
    # four kernels
    fn()

    # no kernels
    g = torch.cuda.CUDAGraph()
    with torch.cuda.graph(g):
        for _ in range(10):
            fn()

    with proton.scope("test0"):
        g.replay()

    proton.finalize()

    with temp_file.open() as f:
        data = json.load(f)

    children = data[0]["children"]
    # metadata scope + kernels + scope_a + scope_b + test0 + scope_d
    assert len(children) == 8
    test0_frame = None
    for child in children:
        if child["frame"]["name"] == "test0":
            test0_frame = child
            break
    assert test0_frame is not None
    capture_at_frame = test0_frame["children"][0]

    foo_test_frame = None
    scope_a_frame = None
    scope_b_frame = None
    scope_d_frame = None
    for child in capture_at_frame["children"]:
        if child["frame"]["name"] == "foo_test":
            foo_test_frame = child
        if child["frame"]["name"] == "scope_a":
            scope_a_frame = child
        if child["frame"]["name"] == "scope_b":
            scope_b_frame = child
        if child["frame"]["name"] == "scope_d":
            scope_d_frame = child
    assert foo_test_frame is not None
    assert foo_test_frame["metrics"]["bytes"] == 160
    assert foo_test_frame["metrics"]["flops"] == 40
    assert foo_test_frame["metrics"]["count"] == 10
    assert scope_a_frame is not None
    assert scope_a_frame["metrics"]["bytes"] == 160
    assert "count" not in scope_a_frame["metrics"]
    assert scope_b_frame is not None
    assert scope_b_frame["metrics"]["sum"] == 40.0
    assert "count" not in scope_b_frame["metrics"]
    assert scope_d_frame is not None
    assert scope_d_frame["metrics"]["vec"] == [0, 10, 20, 30]


@pytest.mark.skipif(not is_cuda(), reason="Only CUDA backend supports metrics profiling in cudagraphs")
def test_tensor_metrics_cudagraph_deactivate(tmp_path: pathlib.Path, device: str):
    stream = torch.cuda.Stream()
    torch.cuda.set_stream(stream)

    def fn(session):
        proton.deactivate(session)
        with proton.scope("scope_b", metrics={"sum": 4}):
            b = torch.ones((2, 2), device=device)
        proton.activate(session)
        c = b * 2  # noqa: F841

    temp_file = tmp_path / "test_tensor_metrics_cudagraph_deactivate.hatchet"
    session = proton.start(str(temp_file.with_suffix("")), context="shadow", hook="triton")

    # warmup
    fn(session)

    # no kernels
    g = torch.cuda.CUDAGraph()
    with torch.cuda.graph(g):
        for _ in range(10):
            fn(session)

    with proton.scope("test0"):
        g.replay()

    proton.finalize()

    # only a single kernel b * 2
    with temp_file.open() as f:
        data = json.load(f)
        children = data[0]["children"]
        test0_frame = None
        for child in children:
            if child["frame"]["name"] == "test0":
                test0_frame = child
                break
        assert test0_frame is not None
        capture_at_frame = test0_frame["children"][0]
        scope_b_frame = None
        c_frame = None
        for child in capture_at_frame["children"]:
            if child["frame"]["name"] == "scope_b":
                scope_b_frame = child
            if "elementwise" in child["frame"]["name"]:
                c_frame = child
        assert scope_b_frame is None
        assert c_frame is not None
        assert c_frame["metrics"]["count"] == 10


@pytest.mark.skipif(not is_cuda(), reason="Only CUDA backend supports metrics profiling in cudagraphs")
def test_tensor_metrics_multi_device_cudagraph(tmp_path: pathlib.Path):
    if torch.cuda.device_count() < 2:
        pytest.skip("Requires at least two CUDA devices")

    devices = [torch.device(f"cuda:{i}") for i in range(2)]
    streams = []
    for device in devices:
        with torch.cuda.device(device):
            streams.append(torch.cuda.Stream(device=device))

    def metadata_fn(grid: tuple, metadata: NamedTuple, args: dict):
        x = args["x"]
        x_sum = x.sum()
        device_idx = x.device.index
        return {"name": f"foo_test_{device_idx}", "bytes": x.numel() * x.element_size(), "flops": x_sum}

    @triton.jit(launch_metadata=metadata_fn)
    def foo(x, y, z):
        tl.store(z, tl.load(y) + tl.load(x))

    def run_on_device(device_id):
        with proton.scope(f"scope_a_{device_id}", metrics={"bytes": 4 * 4}):
            a = torch.ones((2, 2), device=f"cuda:{device_id}")
        with proton.metadata_state():
            a_sum = a.sum()
        with proton.scope(f"scope_b_{device_id}", metrics={"sum": a_sum}):
            b = torch.ones((2, 2), device=f"cuda:{device_id}")
        c = a + b
        foo[(1, )](a, b, c)

    temp_file = tmp_path / "test_tensor_metrics_multi_device_cudagraph.hatchet"
    proton.start(str(temp_file.with_suffix("")), context="shadow", hook="triton")

    graphs = []
    for device, stream in zip(devices, streams):
        with torch.cuda.device(device):
            torch.cuda.set_stream(stream)
            # warmup
            run_on_device(device.index)
            # graph capture
            g = torch.cuda.CUDAGraph()
            with torch.cuda.graph(g, stream=stream):
                for _ in range(10):
                    run_on_device(device.index)
        graphs.append((device, stream, g))

    for device, stream, graph in graphs:
        with torch.cuda.device(device):
            torch.cuda.set_stream(stream)
            with proton.scope(f"test_device_{device.index}"):
                graph.replay()

    proton.finalize()

    with temp_file.open() as f:
        data = json.load(f)

    children = data[0]["children"]
    for device in devices:
        device_name = f"test_device_{device.index}"
        launch_frame = next((child for child in children if child["frame"]["name"] == device_name), None)
        assert launch_frame is not None
        capture_at_frame = launch_frame["children"][0]
        assert capture_at_frame["frame"]["name"] == "<captured_at>"

        foo_frame = None
        scope_a_frame = None
        scope_b_frame = None
        for child in capture_at_frame["children"]:
            if child["frame"]["name"] == f"foo_test_{device.index}":
                foo_frame = child
            if child["frame"]["name"] == f"scope_a_{device.index}":
                scope_a_frame = child
            if child["frame"]["name"] == f"scope_b_{device.index}":
                scope_b_frame = child

        assert foo_frame is not None
        assert scope_a_frame is not None
        assert scope_b_frame is not None
        assert foo_frame["metrics"]["bytes"] == 160
        assert foo_frame["metrics"]["flops"] == 40
        assert foo_frame["metrics"]["device_id"] == str(device.index)
        assert scope_a_frame["metrics"]["bytes"] == 160
        assert scope_b_frame["metrics"]["sum"] == 40.0

    assert len(data) > 1
    cuda_devices = data[1].get("CUDA", {})
    assert len(cuda_devices) >= 2


@pytest.mark.parametrize("buffer_size", [256 * 1024, 64 * 1024 * 1024])
@pytest.mark.parametrize("data_format", ["hatchet_msgpack", "hatchet"])
def test_periodic_flushing(tmp_path, fresh_knobs, data_format, buffer_size, device: str):
    fresh_knobs.proton.profile_buffer_size = buffer_size
    temp_file = tmp_path / f"test_periodic_flushing.{data_format}"
    session = proton.start(str(temp_file.with_suffix("")), mode=f"periodic_flushing:format={data_format}")

    for i in range(10000):
        if i != 0 and i % 1000 == 0:
            proton.data.advance_phase(session=session)
        with proton.scope(f"test_{i}", metrics={"count": 1}):
            torch.zeros((100), device=device)

    proton.finalize(output_format=data_format)

    # Find all *.hatchet files under the directory `tmp_path`
    import glob
    import msgpack
    hatchet_files = glob.glob(str(tmp_path / f"*.{data_format}"))
    assert len(hatchet_files) == 10
    num_scopes = 0
    for hatchet_file in hatchet_files:
        if data_format == "hatchet_msgpack":
            with open(hatchet_file, "rb") as f:
                data = msgpack.load(f, raw=False, strict_map_key=False)
        else:
            with open(hatchet_file, "r", encoding="utf-8") as f:
                data = json.load(f)
        assert len(data[0]["children"]) == 1000
        assert data[0]["children"][0]["metrics"]["count"] == 1
        assert data[0]["children"][0]["frame"]["name"].startswith("test_")
        assert data[0]["children"][0]["children"][0]["metrics"]["time (ns)"] > 0
        num_scopes += len(data[0]["children"])
    assert num_scopes == 10000


@pytest.mark.skipif(not is_cuda(), reason="Only CUDA backend supports metrics profiling in cudagraphs")
@pytest.mark.parametrize("buffer_size", [256 * 1024, 64 * 1024 * 1024])
@pytest.mark.parametrize("data_format", ["hatchet_msgpack", "hatchet"])
def test_periodic_flushing_cudagraph(tmp_path, fresh_knobs, data_format, buffer_size, device: str):
    fresh_knobs.proton.profile_buffer_size = buffer_size
    temp_file = tmp_path / f"test_periodic_flushing.{data_format}"
    session = proton.start(str(temp_file.with_suffix("")), mode=f"periodic_flushing:format={data_format}",
                           hook="triton")

    def metadata_fn(grid: tuple, metadata: NamedTuple, args: dict):
        x = args["x"]
        x_sum = x.sum()
        return {"name": "foo_test", "bytes": x.numel() * x.element_size(), "flops": x_sum}

    @triton.jit(launch_metadata=metadata_fn)
    def foo(x, y, z):
        tl.store(z, tl.load(y) + tl.load(x))

    def fn():
        with proton.scope("scope_a", metrics={"bytes": 4 * 4}):
            a = torch.ones((2, 2), device=device)
        c = a + a
        foo[(1, )](a, a, c)

    # warmup
    fn()

    # no kernels
    g = torch.cuda.CUDAGraph()
    with torch.cuda.graph(g):
        fn()

    with proton.scope("test0"):
        for i in range(10000):
            if i != 0 and i % 1000 == 0:
                proton.data.advance_phase(session=session)
            g.replay()

    proton.finalize(output_format=data_format)

    # Find all *.hatchet files under the directory `tmp_path`
    import glob
    import msgpack
    hatchet_files = glob.glob(str(tmp_path / f"*.{data_format}"))
    assert len(hatchet_files) == 10
    for hatchet_file in hatchet_files:
        if data_format == "hatchet_msgpack":
            with open(hatchet_file, "rb") as f:
                data = msgpack.load(f, raw=False, strict_map_key=False)
        else:
            with open(hatchet_file, "r", encoding="utf-8") as f:
                data = json.load(f)
        capture_frame = None
        for child in data[0]["children"]:
            if child["frame"]["name"] == "test0":
                capture_frame = child["children"][0]
                break
        assert capture_frame is not None
        scope_a_frame = None
        foo_test_frame = None
        for child in capture_frame["children"]:
            if child["frame"]["name"] == "scope_a":
                scope_a_frame = child
            if child["frame"]["name"] == "foo_test":
                foo_test_frame = child
        assert scope_a_frame is not None
        assert foo_test_frame is not None
        assert scope_a_frame["metrics"]["bytes"] == 16000
        assert foo_test_frame["metrics"]["bytes"] == 16000
        assert foo_test_frame["metrics"]["flops"] == 4000


@pytest.mark.skipif(not is_blackwell(), reason="HW trace is only supported on Blackwell GPUs")
def test_hw_trace(fresh_knobs, tmp_path: pathlib.Path, device: str):
    fresh_knobs.proton.enable_hw_trace = True
    temp_file = tmp_path / "test_hw_trace.hatchet"
    proton.start(str(temp_file.with_suffix("")), hook="triton")

    with proton.scope("init"):
        x = torch.ones((1024, ), device=device, dtype=torch.float32)  # noqa: F841

    proton.finalize()

    with temp_file.open() as f:
        data = json.load(f)
    kernel_frame = data[0]["children"][0]["children"][0]
    assert "elementwise" in kernel_frame["frame"]["name"]
    assert kernel_frame["metrics"]["time (ns)"] > 0



import triton.profiler as proton

import torch
import sys

from helper_kernels import custom_add, matmul_kernel


def main():
    a = torch.zeros(1, device="cuda")
    with proton.scope("test"):
        custom_add[(1, )](a)


def test_main():
    main()


def matmul():
    a = torch.randn((32, 32), device="cuda", dtype=torch.float16)
    b = torch.randn((32, 32), device="cuda", dtype=torch.float16)
    M, K = a.shape
    K, N = b.shape
    c = torch.empty((M, N), device=a.device, dtype=a.dtype)

    matmul_kernel[(1, )](
        a, b, c,  #
        M, N, K,  #
        a.stride(0), a.stride(1),  #
        b.stride(0), b.stride(1),  #
        c.stride(0), c.stride(1),  #
        128, 256, 64, 8)
    return c


if __name__ == "__main__":
    if sys.argv[1] == "test":
        main()
    elif sys.argv[1] == "test_matmul":
        matmul()



import pytest
import subprocess
import json
import pathlib


def test_help():
    # Only check if the viewer can be invoked
    subprocess.check_call(["proton", "-h"], stdout=subprocess.DEVNULL)


@pytest.mark.parametrize("mode", ["script", "python", "pytest"])
def test_exec(mode, tmp_path: pathlib.Path):
    file_path = __file__
    helper_file = file_path.replace("test_cmd.py", "helper.py")
    temp_file = tmp_path / "test_exec.hatchet"
    name = str(temp_file.with_suffix(""))
    if mode == "script":
        subprocess.check_call(["proton", "-n", name, helper_file, "test"], stdout=subprocess.DEVNULL)
    elif mode == "python":
        subprocess.check_call(["python3", "-m", "triton.profiler.proton", "-n", name, helper_file, "test"],
                              stdout=subprocess.DEVNULL)
    elif mode == "pytest":
        subprocess.check_call(["proton", "-n", name, "pytest", "-k", "test_main", helper_file],
                              stdout=subprocess.DEVNULL)
    with temp_file.open() as f:
        data = json.load(f, )
    kernels = data[0]["children"]
    assert len(kernels) == 2
    assert kernels[0]["frame"]["name"] == "test" or kernels[1]["frame"]["name"] == "test"



import functools
import triton

from triton._C.libproton import proton as libproton  # type: ignore
from triton._C.libtriton import getenv  # type: ignore
from .flags import flags
from .hooks import HookManager, LaunchHook, InstrumentationHook
from .hooks.hook import Hook
from .mode import BaseMode
from typing import Optional, Union

DEFAULT_PROFILE_NAME = "proton"


def _select_backend() -> str:
    backend = triton.runtime.driver.active.get_current_target().backend
    if backend == "cuda":
        return "cupti"
    elif backend == "hip":
        return "roctracer"
    else:
        raise ValueError("No backend is available for the current target.")


def _get_mode_str(backend: str, mode: Optional[Union[str, BaseMode]]) -> str:
    if backend == "instrumentation":
        prefix = triton.runtime.driver.active.get_current_target().backend
        return f"{prefix}:{mode}" if mode else prefix
    return str(mode) if mode else ""


def _check_env(backend: str) -> None:
    if backend == "roctracer":
        hip_device_envs = ["HIP_VISIBLE_DEVICES", "CUDA_VISIBLE_DEVICES"]
        for env in hip_device_envs:
            if getenv(env, None) is not None:
                raise ValueError(
                    f"Proton does not work when the environment variable {env} is set on AMD GPUs. Please unset it and use `ROCR_VISIBLE_DEVICES` instead"
                )

    use_blackwell_cupti = backend == "cupti" and triton.runtime.driver.active.get_current_target().arch >= 100

    # Ensure default envs are set for Proton knobs if not already set by the user.
    for attr, desc in triton.knobs.proton.knob_descriptors.items():
        key = desc.key
        if getenv(key, None) is None:
            if key == "TRITON_CUPTI_LIB_PATH" and use_blackwell_cupti:
                # For Blackwell+ GPUs, use the cupti library built for Blackwell.
                triton.knobs.setenv(key, triton.knobs.proton.cupti_lib_blackwell_dir)
                continue
            val = getattr(triton.knobs.proton, attr)
            if val is not None:
                if env_val := triton.knobs.toenv(val):
                    triton.knobs.setenv(key, env_val[0])


def start(
    name: Optional[str] = None,
    *,
    context: Optional[str] = "shadow",
    data: Optional[str] = "tree",
    backend: Optional[str] = None,
    mode: Optional[Union[str, BaseMode]] = None,
    hook: Optional[Union[str, Hook]] = None,
) -> Optional[int]:
    """
    Start profiling with the given name and backend.

    Usage:

        ```python
        proton.start("my_profile")
        # do something
        proton.finalize()
        ```

    Args:
        name (str, optional): The name (with path) of the profiling session.
                              If not provided, the default name is "~/proton.<suffix>", where suffix is the default
                              format according to the data type. For example, if data is "tree", the default name is "~/proton.hatchet".
        context (str, optional): The context to use for profiling.
                                 Available options are ["shadow", "python"].
                                 Defaults to "shadow".
        data (str, optional): The data structure to use for profiling.
                              Available options are ["tree", "trace"].
                              Defaults to "tree".
        backend (str, optional): The backend to use for profiling.
                                 Available options are [None, "cupti", "roctracer", "instrumentation"].
                                 Defaults to None, which automatically selects the backend matching the current active runtime.
        mode (Union[str, BaseMode], optional): The "mode" to use for profiling, which is specific to the backend.
                                               Can be a string or an instance of BaseMode (or any subclass thereof).
                                               Defaults to None.
                                               For "cupti", available options are [None, "pcsampling", "periodic_flushing"].
                                               For "roctracer", available options are ["periodic_flushing"].
                                               For "instrumentation", available options are [None].
                                               Each mode has a set of control knobs following with the mode name.
                                               For example, "periodic_flushing" mode has a knob:
                                               - format: The output format of the profiling results. Available options are ["hatchet", "hatchet_msgpack", "chrome_trace"]. Default is "hatchet".
                                               The can be set via `mode="periodic_flushing:format=chrome_trace"`.
        hook (Union[str, Hook], optional): The hook to use for profiling.
                                           You may pass either:
                                           - a string hook name, e.g. "triton" (kernel launch metadata), or
                                           - a custom Hook instance.
                                           Defaults to None.
    Returns:
        session (Optional[int]): The session ID of the profiling session, or None if profiling is disabled.
    """
    if flags.command_line or triton.knobs.proton.disable:
        # Ignore the start() call if the script is run from the command line or profiling is disabled.
        return None

    flags.profiling_on = True

    name = DEFAULT_PROFILE_NAME if name is None else name
    backend = _select_backend() if backend is None else backend
    # Convert mode to its string representation for libproton's runtime
    mode_str = _get_mode_str(backend, mode)

    _check_env(backend)

    session = libproton.start(name, context, data, backend, mode_str)

    if isinstance(hook, Hook):
        HookManager.register(hook, session)
    elif hook == "triton":
        HookManager.register(LaunchHook(), session)
    elif hook is not None:
        raise ValueError(f"Unsupported hook: {hook!r}")
    if backend == "instrumentation":
        HookManager.register(InstrumentationHook(mode), session)

    return session


def activate(session: Optional[int] = None) -> None:
    """
    Activate the specified session.
    The profiling session will be active and data will be recorded.

    Args:
        session (int): The session ID of the profiling session. Defaults to None (all sessions)

    Returns:
        None
    """
    if flags.command_line and session != 0:
        raise ValueError("Only one session can be activated when running from the command line.")

    HookManager.activate(session)

    if session is None:
        libproton.activate_all()
    else:
        libproton.activate(session)


def deactivate(session: Optional[int] = None, flushing: bool = False) -> None:
    """
    Stop the specified session.
    The profiling session's data will still be in the memory, but no more data will be recorded.

    Args:
        session (int): The session ID of the profiling session. Defaults to None (all sessions)
        flushing (bool): Whether to flush the profiling data before deactivating. Defaults to True.

    Returns:
        None
    """
    if flags.command_line and session != 0:
        raise ValueError("Only one session can be deactivated when running from the command line.")

    HookManager.deactivate(session)

    if session is None:
        libproton.deactivate_all(flushing)
    else:
        libproton.deactivate(session, flushing)


def finalize(session: Optional[int] = None, output_format: Optional[str] = "") -> None:
    """
    Finalizes a profiling session.
    Flush and write the profiling data to the file specified by the session name.

    Args:
        session (int, optional): The session ID to finalize. If None, all sessions are finalized. Defaults to None.
        output_format (str, optional): The output format for the profiling results.
                                       Available options are ["hatchet", "hatchet_msgpack", "chrome_trace"].

    Returns:
        None
    """
    HookManager.unregister(session)

    if session is None:
        flags.profiling_on = False
        libproton.finalize_all(output_format)
    else:
        if flags.command_line and session != 0:
            raise ValueError("Only one session can be finalized when running from the command line.")
        libproton.finalize(session, output_format)


def _profiling(
    func,
    name: Optional[str] = None,
    context: Optional[str] = "shadow",
    data: Optional[str] = "tree",
    backend: Optional[str] = None,
    mode: Optional[str] = None,
    hook: Optional[Union[str, Hook]] = None,
):
    """
    Context manager for profiling. Internally use only.

    Args:
        See start() for the arguments.

    Returns:
        wrapper (function): The wrapped function.
    """

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        session = start(name, context=context, data=data, backend=backend, mode=mode, hook=hook)
        ret = func(*args, **kwargs)
        deactivate(session)
        return ret

    return wrapper


def profile(
    func=None,
    *,
    name: Optional[str] = None,
    context: Optional[str] = "shadow",
    data: Optional[str] = "tree",
    backend: Optional[str] = None,
    mode: Optional[str] = None,
    hook: Optional[Union[str, Hook]] = None,
):
    """
    Decorator for profiling.

    Usage:

    ```python
    @proton.profile
    def foo():
        pass
    ```

    Args:
        See start() for the arguments.

    Returns:
        decorator (function): The decorator function.
    """
    if func is None:
        # It's being used with parentheses, so return a decorator
        def decorator(f):
            return _profiling(f, name=name, context=context, data=data, backend=backend, mode=mode, hook=hook)

        return decorator
    else:
        # It's being used without parentheses, so apply the decorator directly
        return _profiling(func, name=name, context=context, data=data, backend=backend, mode=mode, hook=hook)



import triton
import argparse
import ctypes
import triton.profiler as proton
import triton.profiler.language as pl
from triton.profiler.hooks import InstrumentationHook

pl.enable_semantic("triton")


def write_tensor_to_file(tensor, filename):
    data_ptr = tensor.data_ptr()
    size = tensor.numel()
    dtype_size = tensor.element_size()
    total_bytes = size * dtype_size

    with open(filename, 'wb') as f:
        data_arr = ctypes.cast(data_ptr, ctypes.POINTER(ctypes.c_ubyte * total_bytes))
        f.write(bytes(data_arr.contents))


@triton.jit
def seq_kernel():
    pl.enter_scope("r0")
    pl.enter_scope("r1")
    pl.enter_scope("r2")
    pl.exit_scope("r1")
    pl.exit_scope("r0")
    pl.exit_scope("r2")


def seq(args):
    grid_size = 2
    grid = (grid_size, )
    proton.start("", backend="instrumentation", mode=proton.mode.Default(buffer_size=256))
    InstrumentationHook.enable_host_buffer = True
    InstrumentationHook.profile_buffer_size = 512
    seq_kernel[grid]()
    write_tensor_to_file(InstrumentationHook.host_buffer, args.trace_file)
    proton.finalize()


@triton.jit
def loop_kernel():
    for k in range(0, 20):
        pl.enter_scope("r0")
        pl.exit_scope("r0")


def loop(args):
    grid_size = 1
    grid = (grid_size, )
    proton.start("", backend="instrumentation", mode=proton.mode.Default(buffer_size=256))
    InstrumentationHook.enable_host_buffer = True
    InstrumentationHook.profile_buffer_size = 512
    loop_kernel[grid]()
    write_tensor_to_file(InstrumentationHook.host_buffer, args.trace_file)
    proton.finalize()


def main():
    parser = argparse.ArgumentParser(description='Proton intra kernel profiler trace generator')
    parser.add_argument('trace_file', type=str, help='Trace file path')
    parser.add_argument('--kernel', '-k', type=str, help='Kernel name')
    args = parser.parse_args()

    if args.kernel == "seq":
        seq(args)
    if args.kernel == "loop":
        loop(args)


if __name__ == '__main__':
    main()



import pytest


def pytest_addoption(parser):
    parser.addoption("--device", action="store", default="cuda")


@pytest.fixture
def device(request):
    return request.config.getoption("--device")


@pytest.fixture
def fresh_knobs():
    from triton._internal_testing import _fresh_knobs_impl

    fresh_function, reset_function = _fresh_knobs_impl()
    try:
        yield fresh_function()
    finally:
        reset_function()



import torch

import triton
import triton.language as tl
import triton.profiler as proton
from typing import NamedTuple
import argparse


def unpack_grid(grid):
    if len(grid) == 1:
        return grid[0], 1, 1
    if len(grid) == 2:
        return grid[0], grid[1], 1
    if len(grid) == 3:
        return grid[0], grid[1], grid[2]


def metadata_fn(
    grid: tuple,
    metadata: NamedTuple,
    args: dict,
):
    grid_x, grid_y, grid_z = unpack_grid(grid)
    num_warps = metadata.num_warps
    num_stages = metadata.num_stages
    cluster_x, cluster_y, cluster_z = unpack_grid((metadata.num_ctas, ))
    shared_memory = metadata.shared
    M, K = args["a_ptr"].shape
    K, N = args["b_ptr"].shape
    return {
        "name":
        f"matmul_<grid:{grid_x}x{grid_y}x{grid_z}>_<cluster:{cluster_x}x{cluster_y}x{cluster_z}>_<warps:{num_warps}>_<shared:{shared_memory}>_<stages:{num_stages}>",
        "flops": 2 * M * N * K,
        "bytes": (M * N + N * K + K * M) * args["a_ptr"].element_size(),
    }


@triton.autotune(
    configs=[
        triton.Config(
            {
                "BLOCK_SIZE_M": 128,
                "BLOCK_SIZE_N": 256,
                "BLOCK_SIZE_K": 64,
                "GROUP_SIZE_M": 8,
            },
            num_stages=3,
            num_warps=8,
        ),
        triton.Config(
            {
                "BLOCK_SIZE_M": 64,
                "BLOCK_SIZE_N": 256,
                "BLOCK_SIZE_K": 32,
                "GROUP_SIZE_M": 8,
            },
            num_stages=4,
            num_warps=4,
        ),
        triton.Config(
            {
                "BLOCK_SIZE_M": 128,
                "BLOCK_SIZE_N": 128,
                "BLOCK_SIZE_K": 32,
                "GROUP_SIZE_M": 8,
            },
            num_stages=4,
            num_warps=4,
        ),
        triton.Config(
            {
                "BLOCK_SIZE_M": 128,
                "BLOCK_SIZE_N": 64,
                "BLOCK_SIZE_K": 32,
                "GROUP_SIZE_M": 8,
            },
            num_stages=4,
            num_warps=4,
        ),
        triton.Config(
            {
                "BLOCK_SIZE_M": 64,
                "BLOCK_SIZE_N": 128,
                "BLOCK_SIZE_K": 32,
                "GROUP_SIZE_M": 8,
            },
            num_stages=4,
            num_warps=4,
        ),
        triton.Config(
            {
                "BLOCK_SIZE_M": 128,
                "BLOCK_SIZE_N": 32,
                "BLOCK_SIZE_K": 32,
                "GROUP_SIZE_M": 8,
            },
            num_stages=4,
            num_warps=4,
        ),
        triton.Config(
            {
                "BLOCK_SIZE_M": 64,
                "BLOCK_SIZE_N": 32,
                "BLOCK_SIZE_K": 32,
                "GROUP_SIZE_M": 8,
            },
            num_stages=5,
            num_warps=2,
        ),
        triton.Config(
            {
                "BLOCK_SIZE_M": 32,
                "BLOCK_SIZE_N": 64,
                "BLOCK_SIZE_K": 32,
                "GROUP_SIZE_M": 8,
            },
            num_stages=5,
            num_warps=2,
        ),
    ],
    key=["M", "N", "K"],
)
@triton.jit(launch_metadata=metadata_fn)
def matmul_kernel(
        # Pointers to matrices
        a_ptr, b_ptr, c_ptr,
        # Matrix dimensions
        M, N, K,
        # The stride variables represent how much to increase the ptr by when moving by 1
        # element in a particular dimension. E.g. `stride_am` is how much to increase `a_ptr`
        # by to get the element one row down (A has M rows).
        stride_am, stride_ak,  #
        stride_bk, stride_bn,  #
        stride_cm, stride_cn,
        # Meta-parameters
        BLOCK_SIZE_M: tl.constexpr, BLOCK_SIZE_N: tl.constexpr, BLOCK_SIZE_K: tl.constexpr,  #
        GROUP_SIZE_M: tl.constexpr,  #
        ACTIVATION: tl.constexpr,  #
):
    """Kernel for computing the matmul C = A x B.
    A has shape (M, K), B has shape (K, N) and C has shape (M, N)
    """
    # -----------------------------------------------------------
    # Map program ids `pid` to the block of C it should compute.
    # This is done in a grouped ordering to promote L2 data reuse.
    # See above `L2 Cache Optimizations` section for details.
    pid = tl.program_id(axis=0)
    num_pid_m = tl.cdiv(M, BLOCK_SIZE_M)
    num_pid_n = tl.cdiv(N, BLOCK_SIZE_N)
    num_pid_in_group = GROUP_SIZE_M * num_pid_n
    group_id = pid // num_pid_in_group
    first_pid_m = group_id * GROUP_SIZE_M
    group_size_m = min(num_pid_m - first_pid_m, GROUP_SIZE_M)
    pid_m = first_pid_m + (pid % group_size_m)
    pid_n = (pid % num_pid_in_group) // group_size_m

    # ----------------------------------------------------------
    # Create pointers for the first blocks of A and B.
    # We will advance this pointer as we move in the K direction
    # and accumulate
    # `a_ptrs` is a block of [BLOCK_SIZE_M, BLOCK_SIZE_K] pointers
    # `b_ptrs` is a block of [BLOCK_SIZE_K, BLOCK_SIZE_N] pointers
    # See above `Pointer Arithmetic` section for details
    offs_am = (pid_m * BLOCK_SIZE_M + tl.arange(0, BLOCK_SIZE_M)) % M
    offs_bn = (pid_n * BLOCK_SIZE_N + tl.arange(0, BLOCK_SIZE_N)) % N
    offs_k = tl.arange(0, BLOCK_SIZE_K)
    a_ptrs = a_ptr + (offs_am[:, None] * stride_am + offs_k[None, :] * stride_ak)
    b_ptrs = b_ptr + (offs_k[:, None] * stride_bk + offs_bn[None, :] * stride_bn)

    # -----------------------------------------------------------
    # Iterate to compute a block of the C matrix.
    # We accumulate into a `[BLOCK_SIZE_M, BLOCK_SIZE_N]` block
    # of fp32 values for higher accuracy.
    # `accumulator` will be converted back to fp16 after the loop.
    accumulator = tl.zeros((BLOCK_SIZE_M, BLOCK_SIZE_N), dtype=tl.float32)
    for k in range(0, tl.cdiv(K, BLOCK_SIZE_K)):
        # Load the next block of A and B, generate a mask by checking the K dimension.
        # If it is out of bounds, set it to 0.
        a = tl.load(a_ptrs, mask=offs_k[None, :] < K - k * BLOCK_SIZE_K, other=0.0)
        b = tl.load(b_ptrs, mask=offs_k[:, None] < K - k * BLOCK_SIZE_K, other=0.0)
        # We accumulate along the K dimension.
        accumulator += tl.dot(a, b)
        # Advance the ptrs to the next K block.
        a_ptrs += BLOCK_SIZE_K * stride_ak
        b_ptrs += BLOCK_SIZE_K * stride_bk
    # You can fuse arbitrary activation functions here
    # while the accumulator is still in FP32!
    if ACTIVATION == "leaky_relu":
        accumulator = leaky_relu(accumulator)
    c = accumulator.to(tl.float16)

    # -----------------------------------------------------------
    # Write back the block of the output matrix C with masks.
    offs_cm = pid_m * BLOCK_SIZE_M + tl.arange(0, BLOCK_SIZE_M)
    offs_cn = pid_n * BLOCK_SIZE_N + tl.arange(0, BLOCK_SIZE_N)
    c_ptrs = c_ptr + stride_cm * offs_cm[:, None] + stride_cn * offs_cn[None, :]
    c_mask = (offs_cm[:, None] < M) & (offs_cn[None, :] < N)
    tl.store(c_ptrs, c, mask=c_mask)


# We can fuse `leaky_relu` by providing it as an `ACTIVATION` meta-parameter in `_matmul`.
@triton.jit
def leaky_relu(x):
    x = x + 1
    return tl.where(x >= 0, x, 0.01 * x)


# %%
# We can now create a convenience wrapper function that only takes two input tensors,
# and (1) checks any shape constraint; (2) allocates the output; (3) launches the above kernel.


def matmul(a, b, activation=""):
    # Check constraints.
    assert a.shape[1] == b.shape[0], "Incompatible dimensions"
    assert a.is_contiguous(), "Matrix A must be contiguous"
    assert b.is_contiguous(), "Matrix B must be contiguous"
    M, K = a.shape
    K, N = b.shape
    # Allocates output.
    c = torch.empty((M, N), device=a.device, dtype=a.dtype)

    # 1D launch kernel where each block gets its own program.
    def grid(META):
        return (triton.cdiv(M, META["BLOCK_SIZE_M"]) * triton.cdiv(N, META["BLOCK_SIZE_N"]), )

    matmul_kernel[grid](
        a, b, c,  #
        M, N, K,  #
        a.stride(0), a.stride(1),  #
        b.stride(0), b.stride(1),  #
        c.stride(0), c.stride(1),  #
        ACTIVATION=activation,  #
    )
    return c


argparser = argparse.ArgumentParser()
argparser.add_argument("--profile", action="store_true")
argparser.add_argument("--pcsampling", action="store_true", default=False)
argparser.add_argument("--cudagraph", action="store_true", default=False)
args = argparser.parse_args()


@triton.testing.perf_report(
    triton.testing.Benchmark(
        x_names=["M", "N", "K"],  # Argument names to use as an x-axis for the plot
        x_vals=[128 * i for i in range(2, 10)],  # Different possible values for `x_name`
        line_arg="provider",  # Argument name whose value corresponds to a different line in the plot
        # Possible values for `line_arg`
        line_vals=["cublas", "triton"],
        # Label name for the lines
        line_names=["cuBLAS", "Triton"],
        # Line styles
        styles=[("green", "-"), ("blue", "-")],
        ylabel="TFLOPS",  # Label name for the y-axis
        plot_name="matmul-performance",  # Name for the plot, used also as a file name for saving the plot.
        args={},
    ))
def benchmark(M, N, K, provider):
    a = torch.randn((M, K), device="cuda", dtype=torch.float16)
    b = torch.randn((K, N), device="cuda", dtype=torch.float16)
    quantiles = [0.5, 0.2, 0.8]
    with proton.scope(f"matmul_{M}_{N}_{K}"):
        if provider == "cublas":

            @proton.scope(
                "cublas",
                metrics={
                    "flops": 2 * M * N * K,
                    "bytes": (M * N + N * K + K * M) * a.element_size(),
                },
            )
            def cublas_matmul(a, b):
                torch.matmul(a, b)

            if args.cudagraph:
                ms = triton.testing.do_bench_cudagraph(lambda: cublas_matmul(a, b))
                min_ms = max_ms = ms
            else:
                ms, min_ms, max_ms = triton.testing.do_bench(lambda: cublas_matmul(a, b), quantiles=quantiles)
        if provider == "triton":

            def enter_autotune(args, reset_only=False):
                if reset_only:
                    return
                proton.enter_scope("<autotune>")

            def exit_autotune(args, exception):
                proton.exit_scope()

            matmul_kernel.pre_hook = enter_autotune
            matmul_kernel.post_hook = exit_autotune
            with proton.scope("triton"):
                if args.cudagraph:
                    ms = triton.testing.do_bench_cudagraph(lambda: matmul(a, b))
                    min_ms = max_ms = ms
                else:
                    ms, min_ms, max_ms = triton.testing.do_bench(lambda: matmul(a, b), quantiles=quantiles)

    def perf(ms):
        return 2 * M * N * K * 1e-12 / (ms * 1e-3)

    return perf(ms), perf(max_ms), perf(min_ms)


if args.profile:
    if args.pcsampling:
        # proton-viewer -m num_samples/%,time/s ./matmul.hatchet
        proton.start("matmul", hook="triton", backend="cupti", mode="pcsampling")
    else:
        # proton-viewer -m tflop/s,time/s ./matmul.hatchet
        proton.start("matmul", hook="triton")
    benchmark.run(show_plots=True, print_data=True)
    proton.finalize()
else:
    benchmark.run(show_plots=True, print_data=True)



import torch

import triton
import triton.language as tl
import triton.profiler as proton
import pathlib
import sys

from typing import NamedTuple

DEVICE = triton.runtime.driver.active.get_active_torch_device()


def metadata_fn(grid: tuple, metadata: NamedTuple, args: dict):
    BLOCK_SIZE = args["BLOCK_SIZE"]
    return {"name": f"add_{BLOCK_SIZE}"}


@triton.jit(launch_metadata=metadata_fn)
def add_kernel(x_ptr,  # *Pointer* to first input vector.
               y_ptr,  # *Pointer* to second input vector.
               output_ptr,  # *Pointer* to output vector.
               n_elements,  # Size of the vector.
               BLOCK_SIZE: tl.constexpr,  # Number of elements each program should process.
               # NOTE: `constexpr` so it can be used as a shape value.
               ):
    pid = tl.program_id(axis=0)
    block_start = pid * BLOCK_SIZE
    offsets = block_start + tl.arange(0, BLOCK_SIZE)
    mask = offsets < n_elements
    x = tl.load(x_ptr + offsets, mask=mask)
    y = tl.load(y_ptr + offsets, mask=mask)
    output = x + y
    tl.store(output_ptr + offsets, output, mask=mask)


def add(x: torch.Tensor, y: torch.Tensor, path):
    output = torch.empty_like(x)
    assert x.device == DEVICE and y.device == DEVICE and output.device == DEVICE
    n_elements = output.numel()
    grid = lambda meta: (triton.cdiv(n_elements, meta['BLOCK_SIZE']), )
    tmp_path = pathlib.Path(path)
    temp_file = tmp_path / "test_override.hatchet"
    proton.start(str(temp_file.with_suffix("")), backend="instrumentation")
    add_kernel[grid](x, y, output, n_elements, BLOCK_SIZE=1024, num_warps=1)
    proton.finalize()
    return output


size = 98432
x = torch.rand(size, device=DEVICE)
y = torch.rand(size, device=DEVICE)
output_torch = x + y
output_triton = add(x, y, sys.argv[-1])



"""
Vector Addition with Triton Intra-Kernel Profiling using TTGIR Override

This tutorial demonstrates how to use Triton's TTGIR override mechanism
to enable intra-kernel profiling with Proton. The workflow involves generating,
modifying, and overriding the kernel's intermediate representation to insert
profiling hooks.

Workflow:
1. Generate TTGIR dump files:

   This creates the original TTGIR files in the `ttgir_dump/` directory:

   ../../scripts/dump_ttgir.sh python3 example_override.py --increase-accuracy

2. Insert profiling instrumentation:

   Modify the generated TTGIR files by adding proton.record operators at desired
   profiling points. Example script that adds proton ops in the above ttgir:

   ./insert_proton_records

3. Execute with TTGIR override:

   TRITON_ALWAYS_COMPILE=1 TRITON_KERNEL_OVERRIDE=1 TRITON_OVERRIDE_DIR=ttgir_dump python3 example_override.py --increase-accuracy

   - TRITON_ALWAYS_COMPILE=1: Forces recompilation on each run
   - TRITON_KERNEL_OVERRIDE=1: Enables TTGIR override mechanism
   - TRITON_OVERRIDE_DIR=ttgir_dump: Specifies directory containing modified TTGIR files
"""

import argparse

import torch
import triton
import triton.language as tl
import triton.profiler as proton
from triton.profiler.mode import Default

DEVICE = triton.runtime.driver.active.get_active_torch_device()


@triton.jit
def add_kernel(x_ptr,  # *Pointer* to first input vector.
               y_ptr,  # *Pointer* to second input vector.
               output_ptr,  # *Pointer* to output vector.
               n_elements,  # Size of the vector.
               BLOCK_SIZE: tl.constexpr,  # Number of elements each program should process.
               # NOTE: `constexpr` so it can be used as a shape value.
               ):
    pid = tl.program_id(axis=0)
    block_start = pid * BLOCK_SIZE
    offsets = block_start + tl.arange(0, BLOCK_SIZE)
    mask = offsets < n_elements
    x = tl.load(x_ptr + offsets, mask=mask)
    y = tl.load(y_ptr + offsets, mask=mask)
    output = x + y
    tl.store(output_ptr + offsets, output, mask=mask)


def add(x: torch.Tensor, y: torch.Tensor):
    parser = argparse.ArgumentParser(description="TTGIR override example with Triton intra kernel profiling")
    parser.add_argument(
        "--increase-accuracy",
        action="store_true",
        default=False,
        help="Enable increased-accuracy during profiling (default: False)",
    )
    args = parser.parse_args()

    output = torch.empty_like(x)
    assert x.device == DEVICE and y.device == DEVICE and output.device == DEVICE
    n_elements = output.numel()
    grid = lambda meta: (triton.cdiv(n_elements, meta["BLOCK_SIZE"]), )

    if args.increase_accuracy:
        proton.start(
            "add",
            data="trace",
            backend="instrumentation",
            mode=Default(optimizations="clock32,time_shift"),
        )
    else:
        proton.start("add", data="trace", backend="instrumentation")

    add_kernel[grid](x, y, output, n_elements, BLOCK_SIZE=1024)

    proton.finalize()
    return output


torch.manual_seed(0)
size = 98432
x = torch.rand(size, device=DEVICE)
y = torch.rand(size, device=DEVICE)
output_torch = x + y
output_triton = add(x, y)
torch.testing.assert_close(output_torch, output_triton, rtol=1e-3, atol=1e-1)



"""
Centralized, process-local flags with a minimal interface (no environment variables).

Usage:
    from triton.profiler.flags import flags

    # Toggle
    flags.profiling_on = True
    flags.instrumentation_on = False

    # Check
    if flags.command_line:
            ...
"""
from dataclasses import dataclass


@dataclass
class ProfilerFlags:
    # Whether profiling is enabled. Default is False.
    profiling_on: bool = False
    # Whether instrumentation is enabled. Default is False.
    instrumentation_on: bool = False
    # Whether the script is run from the command line. Default is False.
    command_line: bool = False


flags = ProfilerFlags()
